from enum import IntEnum


class ThumbALUOpcode(IntEnum):
    AND = 0x0
    EOR = 0x1
    LSL = 0x2
    LSR = 0x3
    ASR = 0x4
    ADC = 0x5
    SBC = 0x6
    ROR = 0x7
    TST = 0x8
    NEG = 0x9
    CMP = 0xA
    CMN = 0xB
    ORR = 0xC
    MUL = 0xD
    BIC = 0xE
    MVN = 0xF



def thumb_multiple_load_store(cpu: CPU, instr: int):
    """Execute a THUMB.15 instruction (multiple load/store)"""
    reg_list = get_bits(instr, 0, 7)
    base_reg = get_bits(instr, 8, 10)
    address = cpu.regs.get(base_reg)
    cpu.advance_pc_thumb()
    load = get_bit(instr, 11)
    if reg_list == 0:
        if load:
            cpu.regs.set_pc(cpu.memory.read_32(cpu.regs.get(base_reg), MemoryAccess.NON_SEQUENTIAL))
            cpu.flush_pipeline()
        else:
            cpu.memory.write_32(cpu.regs.get(base_reg), cpu.regs.get_pc(), MemoryAccess.NON_SEQUENTIAL)
        cpu.regs.set(base_reg, add_32(cpu.regs.get(base_reg), 64))
        return
    count = 0
    for reg in range(8):
        count += get_bit(instr, reg)
    count *= 4
    access_type = MemoryAccess.NON_SEQUENTIAL
    if load:
        cpu.regs.set(base_reg, add_32(cpu.regs.get(base_reg), count))
        cpu.scheduler.idle(1)
        for reg in range(8):
            if get_bit(reg_list, reg):
                cpu.regs.set(reg, cpu.memory.read_32(address, access_type))
                address = add_32(address, 4)
                access_type = MemoryAccess.SEQUENTIAL
    else:
        first = True
        for reg in range(8):
            if get_bit(reg_list, reg):
                cpu.memory.write_32(address, cpu.regs.get(reg), access_type)
                address = add_32(address, 4)
                access_type = MemoryAccess.SEQUENTIAL
                if first:
                    cpu.regs.set(base_reg, add_32(cpu.regs.get(base_reg), count))
                    first = False

def thumb_push_pop_registers(cpu: CPU, instr: int):
    """Execute a THUMB.14 instruction (push/pop registers to the stack)"""
    reg_list = get_bits(instr, 0, 8)
    opcode = get_bit(instr, 11)
    cpu.advance_pc_thumb()
    if opcode:
        thumb_pop_registers(cpu, reg_list)
    else:
        thumb_push_registers(cpu, reg_list)

def thumb_push_registers(cpu: CPU, reg_list: int):
    """
    Push the specified registers to the stack.

    :param cpu:      The CPU executing this instruction.
    :param reg_list: A 9-bit wide bitfield specifying which registers to push.
                     Bits 0-7 correspond to registers R0-R7 and bit 8 corresponds
                     to LR (R14).
    """
    if reg_list == 0:
        cpu.regs.set_sp(add_32(cpu.regs.get_sp(), -64))
        cpu.memory.write_32(cpu.regs.get_sp(), cpu.regs.get_pc(), MemoryAccess.NON_SEQUENTIAL)
        return
    push_lr = get_bit(reg_list, 8)
    if push_lr:
        cpu.regs.set_sp(add_32(cpu.regs.get_sp(), -4))
        cpu.memory.write_32(cpu.regs.get_sp(), cpu.regs.get_lr(), MemoryAccess.NON_SEQUENTIAL)
    for reg in range(7, -1, -1):
        if get_bit(reg_list, reg):
            cpu.regs.set_sp(add_32(cpu.regs.get_sp(), -4))
            cpu.memory.write_32(cpu.regs.get_sp(), cpu.regs.get(reg), MemoryAccess.SEQUENTIAL)

def thumb_pop_registers(cpu: CPU, reg_list: int):
    """
    Pop the specified registers off the stack.

    :param cpu:      The CPU executing this instruction.
    :param reg_list: A 9-bit wide bitfield specifying which registers to pop.
                     Bits 0-7 correspond to registers R0-R7 and bit 8 corresponds
                     to PC (R15).
    """
    if reg_list == 0:
        cpu.regs.set_pc(cpu.memory.read_32(cpu.regs.get_sp(), MemoryAccess.NON_SEQUENTIAL))
        cpu.regs.set_sp(add_32(cpu.regs.get_sp(), 64))
        cpu.flush_pipeline()
        return
    access_type = MemoryAccess.NON_SEQUENTIAL
    for reg in range(8):
        if get_bit(reg_list, reg):
            cpu.regs.set(reg, cpu.memory.read_32(cpu.regs.get_sp(), access_type))
            cpu.regs.set_sp(add_32(cpu.regs.get_sp(), 4))
            access_type = MemoryAccess.SEQUENTIAL
    pop_pc = get_bit(reg_list, 8)
    if pop_pc:
        cpu.regs.set_pc(cpu.memory.read_32(cpu.regs.get_sp(), access_type))
        cpu.regs.set_sp(add_32(cpu.regs.get_sp(), 4))
        cpu.flush_pipeline()
    cpu.scheduler.idle(1)


from enum import IntEnum

class ThumbALUOpcode(IntEnum):
    AND = 0
    EOR = 1
    LSL = 2
    LSR = 3
    ASR = 4
    ADC = 5
    SBC = 6
    ROR = 7
    TST = 8
    NEG = 9
    CMP = 10
    CMN = 11
    ORR = 12
    MUL = 13
    BIC = 14
    MVN = 15


from libc.stdint cimport uint8_t, uint32_t
from cpython.array cimport array

from pyboy_advance.memory.constants cimport MemoryRegion
from pyboy_advance.ppu.constants cimport VideoMode
from pyboy_advance.ppu.registers cimport DisplayControlRegister
from pyboy_advance.utils cimport (
    get_bit,
    array_read_32,
    array_read_16,
    array_write_32,
    array_write_16,
)

cdef class VideoMemory:
    cdef DisplayControlRegister display_control
    cdef uint8_t[:] palram
    cdef uint8_t[:] vram
    cdef uint8_t[:] oam

    cdef uint32_t read_32_palram(self, uint32_t) noexcept
    cdef uint32_t read_32_vram(self, uint32_t) noexcept
    cdef uint32_t read_32_oam(self, uint32_t) noexcept
    cdef uint32_t read_16_palram(self, uint32_t) noexcept
    cdef uint32_t read_16_vram(self, uint32_t) noexcept
    cdef uint32_t read_16_oam(self, uint32_t) noexcept
    cdef uint32_t read_8_palram(self, uint32_t) noexcept
    cdef uint32_t read_8_vram(self, uint32_t) noexcept
    cdef uint32_t read_8_oam(self, uint32_t) noexcept
    cdef void write_32_palram(self, uint32_t, uint32_t) noexcept
    cdef void write_32_vram(self, uint32_t, uint32_t) noexcept
    cdef void write_32_oam(self, uint32_t, uint32_t) noexcept
    cdef void write_16_palram(self, uint32_t, uint32_t) noexcept
    cdef void write_16_vram(self, uint32_t, uint32_t) noexcept
    cdef void write_16_oam(self, uint32_t, uint32_t) noexcept
    cdef void write_8_palram(self, uint32_t, uint32_t) noexcept
    cdef void write_8_vram(self, uint32_t, uint32_t) noexcept
    cdef uint32_t _get_vram_address_mask(self, uint32_t) noexcept



import pytest

from pyboy_advance.constants import Key


@pytest.mark.rom_url(
    "https://raw.githubusercontent.com/destoer/armwrestler-gba-fixed/master/armwrestler-gba-fixed.gba"
)
def test_armwrestler_arm(emulator, assert_expected_image):
    emulator.frame(20)

    expected_images = [
        "armwrestler_arm_alu_1.png",
        "armwrestler_arm_alu_2.png",
        "armwrestler_arm_load_1.png",
        "armwrestler_arm_load_2.png",
        "armwrestler_arm_ldm_stm.png",
    ]

    for expected_image in expected_images:
        emulator.press_key(Key.BUTTON_START)
        emulator.frame(20)
        emulator.release_key(Key.BUTTON_START)
        emulator.frame(20)
        assert_expected_image(emulator.screen.image, expected_image)


@pytest.mark.rom_url(
    "https://raw.githubusercontent.com/destoer/armwrestler-gba-fixed/master/armwrestler-gba-fixed.gba"
)
def test_armwrestler_thumb(emulator, assert_expected_image):
    emulator.frame(20)

    for _ in range(3):
        emulator.press_key(Key.DPAD_DOWN)
        emulator.frame(20)
        emulator.release_key(Key.DPAD_DOWN)
        emulator.frame(20)

    expected_images = [
        "armwrestler_thumb_alu.png",
        "armwrestler_thumb_ldr_str.png",
        "armwrestler_thumb_ldm_stm.png",
    ]

    for expected_image in expected_images:
        emulator.press_key(Key.BUTTON_START)
        emulator.frame(20)
        emulator.release_key(Key.BUTTON_START)
        emulator.frame(20)
        assert_expected_image(emulator.screen.image, expected_image)



def arm_branch(cpu: CPU, instr: int):
    """Execute a Branch or Branch and Link instruction"""
    offset = extend_sign_24(instr & 16777215) * 4
    link_flag = get_bit(instr, 24)
    if link_flag:
        cpu.regs.set_lr(add_32(cpu.regs.get_pc(), -4))
    cpu.regs.set_pc(add_32(cpu.regs.get_pc(), offset))
    cpu.flush_pipeline()

def arm_branch_exchange(cpu: CPU, instr: int):
    """Execute a Branch and Exchange instruction"""
    address = cpu.regs.get(instr & 15)
    cpu.regs.set_pc(address & ~1)
    cpu.regs.cpsr.set_state(get_bit(address, 0))
    cpu.flush_pipeline()


def arm_single_data_swap(cpu: CPU, instr: int):
    """Executes a Single Data Swap instruction (SWP)"""
    base_reg = get_bits(instr, 16, 19)
    dst_reg = get_bits(instr, 12, 15)
    src_reg = get_bits(instr, 0, 3)
    byte_word_bit = get_bit(instr, 22)
    if byte_word_bit:
        temp = cpu.memory.read_8(cpu.regs.get(base_reg), MemoryAccess.NON_SEQUENTIAL)
        cpu.memory.write_8(cpu.regs.get(base_reg), cpu.regs.get(src_reg), MemoryAccess.NON_SEQUENTIAL)
        cpu.regs.set(dst_reg, temp)
    else:
        temp = cpu.memory.read_32_ror(cpu.regs.get(base_reg), MemoryAccess.NON_SEQUENTIAL)
        cpu.memory.write_32(cpu.regs.get(base_reg), cpu.regs.get(src_reg), MemoryAccess.NON_SEQUENTIAL)
        cpu.regs.set(dst_reg, temp)
    cpu.scheduler.idle(1)
    cpu.advance_pc_arm()
    cpu.next_fetch_access = MemoryAccess.NON_SEQUENTIAL


from libc.stdint cimport uint32_t

from pyboy_advance.cpu.cpu cimport CPU
from pyboy_advance.memory.constants cimport MemoryAccess
from pyboy_advance.utils cimport get_bits, get_bit, add_32

cdef void thumb_pc_relative_load(CPU, uint32_t) noexcept
cdef void thumb_load_store_sign_extended(CPU, uint32_t) noexcept
cdef void thumb_load_store_register_offset(CPU, uint32_t) noexcept
cdef void thumb_load_store_immediate_offset(CPU, uint32_t) noexcept
cdef void thumb_load_store_halfword(CPU, uint32_t) noexcept
cdef void thumb_sp_relative_load_store(CPU, uint32_t) noexcept



ARM_PATTERNS: list[InstrPattern] = [InstrPattern(3840, 3840, arm_software_interrupt), InstrPattern(3584, 2560, arm_branch), InstrPattern(3584, 2048, arm_block_data_transfer), InstrPattern(3072, 1024, arm_single_data_transfer), InstrPattern(4095, 289, arm_branch_exchange), InstrPattern(4031, 265, arm_single_data_swap), InstrPattern(3983, 137, arm_multiply_long), InstrPattern(4047, 9, arm_multiply), InstrPattern(3593, 9, arm_halfword_data_transfer), InstrPattern(3504, 256, arm_mrs), InstrPattern(3504, 288, arm_msr), InstrPattern(3072, 0, arm_alu)]

def fill_arm_lut():
    for lut_index in range(2 ** 12):
        for pattern in ARM_PATTERNS:
            if lut_index & pattern.mask == pattern.value:
                ARM_LUT[lut_index] = pattern.handler
                break
fill_arm_lut()

def arm_decoder(instr: int) -> InstrHandler:
    opcode_hash = instr >> 16 & 4080 | instr >> 4 & 15
    instruction_handler = ARM_LUT[opcode_hash]
    return instruction_handler


# ifndef CYTHON
from __future__ import annotations

from pyboy_advance.ppu.constants import LayerType, WindowIndex, VideoMode, BlendMode
from pyboy_advance.utils import get_bits, get_bit, set_bit, bint
# endif


class DisplayControlRegister:
    def __init__(self):
        self.reg = 0

    @property
    def video_mode(self) -> VideoMode | int:
        return get_bits(self.reg, 0, 2)

    @property
    def page_select(self) -> bint:
        return get_bit(self.reg, 4)

    @property
    def hblank_interval_free(self) -> bint:
        return get_bit(self.reg, 5)

    @property
    def obj_vram_dimension(self) -> bint:
        return get_bit(self.reg, 6)

    @property
    def force_blank(self) -> bint:
        return get_bit(self.reg, 7)

    def display_bg(self, bg_num: int) -> bint:
        return get_bit(self.reg, 8 + bg_num)

    @property
    def display_obj(self) -> bint:
        return get_bit(self.reg, 12)

    def display_window(self, window: WindowIndex) -> bint:
        return get_bit(self.reg, 13 + window)


class DisplayStatusRegister:
    def __init__(self):
        self.reg = 0

    @property
    def vblank_status(self) -> bint:
        return get_bit(self.reg, 0)

    @vblank_status.setter
    def vblank_status(self, value: bint):
        self.reg = set_bit(self.reg, 0, value)

    @property
    def hblank_status(self) -> bint:
        return get_bit(self.reg, 1)

    @hblank_status.setter
    def hblank_status(self, value: bint):
        self.reg = set_bit(self.reg, 1, value)

    @property
    def vcount_trigger_status(self) -> bint:
        return get_bit(self.reg, 2)

    @vcount_trigger_status.setter
    def vcount_trigger_status(self, value: bint):
        self.reg = set_bit(self.reg, 2, value)

    @property
    def vblank_irq(self) -> bint:
        return get_bit(self.reg, 3)

    @vblank_irq.setter
    def vblank_irq(self, value: bint):
        self.reg = set_bit(self.reg, 3, value)

    @property
    def hblank_irq(self) -> bint:
        return get_bit(self.reg, 4)

    @hblank_irq.setter
    def hblank_irq(self, value: bint):
        self.reg = set_bit(self.reg, 4, value)

    @property
    def vcount_irq(self) -> bint:
        return get_bit(self.reg, 5)

    @vcount_irq.setter
    def vcount_irq(self, value: bint):
        self.reg = set_bit(self.reg, 5, value)

    @property
    def vcount_trigger_value(self) -> int:
        return get_bits(self.reg, 8, 15)

    @vcount_trigger_value.setter
    def vcount_trigger_value(self, value: int):
        self.reg = (self.reg & 0xFF) | (value << 8)


class BackgroundControlRegister:
    def __init__(self):
        self.reg = 0

    @property
    def priority(self) -> int:
        return get_bits(self.reg, 0, 1)

    @property
    def tile_set_block(self) -> int:
        return get_bits(self.reg, 2, 3)

    @property
    def mosaic(self) -> bint:
        return get_bit(self.reg, 6)

    @property
    def colour_256(self) -> bint:
        return get_bit(self.reg, 7)

    @property
    def tile_map_block(self) -> int:
        return get_bits(self.reg, 8, 12)

    @property
    def wraparound(self) -> bint:
        return get_bit(self.reg, 13)

    @property
    def size(self) -> int:
        return get_bits(self.reg, 14, 15)


class WindowControlRegister:
    def __init__(self):
        self.reg = 0

    def display_layer(self, layer_type: LayerType) -> bint:
        return get_bit(self.reg, layer_type)

    @property
    def enable_blending(self) -> bint:
        return get_bit(self.reg, 5)


class BlendControlRegister:
    def __init__(self):
        self.reg = 0

    def blend_source(self, layer_type: LayerType | int) -> bint:
        return get_bit(self.reg, layer_type)

    def blend_target(self, layer_type: LayerType | int) -> bint:
        return get_bit(self.reg, 8 + layer_type)

    @property
    def blend_mode(self) -> BlendMode | int:
        return get_bits(self.reg, 6, 7)


class BlendAlphaRegister:
    def __init__(self):
        self.reg = 0

    @property
    def coefficient_source(self) -> int:
        return get_bits(self.reg, 0, 4)

    @property
    def coefficient_target(self) -> int:
        return get_bits(self.reg, 8, 12)


class BlendBrightnessRegister:
    def __init__(self):
        self.reg = 0

    @property
    def brightness(self) -> int:
        return get_bits(self.reg, 0, 4)



from libc.stdint cimport uint32_t

from pyboy_advance.app.constants cimport WindowEvent
from pyboy_advance.constants cimport Interrupt, Key
from pyboy_advance.interrupt_controller cimport InterruptController
from pyboy_advance.utils cimport get_bit, get_bits

cdef class Keypad:
    cdef InterruptController interrupt_controller
    cdef KeypadInterruptControlRegister key_control
    cdef uint32_t key_input

    cdef void press_key(self, int) noexcept
    cdef void release_key(self, int) noexcept
    cdef void process_window_event(self, int) noexcept
    cdef bint _evaluate_irq_condition(self) noexcept

cdef class KeypadInterruptControlRegister:
    cdef uint32_t reg

    cdef uint32_t get_key_select(self) noexcept
    cdef bint get_irq_enable(self) noexcept
    cdef bint get_irq_if_all(self) noexcept



class IO:

    def __init__(self, memory: Memory, interrupt_controller: InterruptController, dma_controller: DMAController, timers: Timers, ppu: PPU, keypad: Keypad):
        self.memory = memory
        self.interrupt_controller = interrupt_controller
        self.dma_controller = dma_controller
        self.timers = timers
        self.ppu = ppu
        self.keypad = keypad
        self.reg_soundbias = 0

    def read_32(self, address: int) -> int:
        lower_bits = self.read_16(address)
        upper_bits = self.read_16(address + 2)
        return upper_bits << 16 | lower_bits

    def read_16(self, address: int) -> int:
        if address == IOAddress.REG_DISPCNT:
            return self.ppu.display_control.reg
        elif address == IOAddress.REG_DISPSTAT:
            return self.ppu.display_status.reg
        elif address == IOAddress.REG_VCOUNT:
            return self.ppu.vcount
        elif address == IOAddress.REG_BG0CNT:
            return self.ppu.bg_control_0.reg
        elif address == IOAddress.REG_BG1CNT:
            return self.ppu.bg_control_1.reg
        elif address == IOAddress.REG_BG2CNT:
            return self.ppu.bg_control_2.reg
        elif address == IOAddress.REG_BG3CNT:
            return self.ppu.bg_control_3.reg
        elif address == IOAddress.REG_WININ:
            win_in = self.ppu.window_control_0.reg
            win_in |= self.ppu.window_control_1.reg << 8
            return win_in
        elif address == IOAddress.REG_WINOUT:
            win_out = self.ppu.window_control_out.reg
            win_out |= self.ppu.window_control_obj.reg << 8
            return win_out
        elif address == IOAddress.REG_BLDCNT:
            return self.ppu.blend_control.reg
        elif address == IOAddress.REG_BLDALPHA:
            return self.ppu.blend_alpha.reg
        elif address == IOAddress.REG_SOUNDBIAS:
            return self.reg_soundbias
        elif address == IOAddress.REG_DMA0CNT_H:
            return self.dma_controller.channel_0.get_control_reg()
        elif address == IOAddress.REG_DMA1CNT_H:
            return self.dma_controller.channel_1.get_control_reg()
        elif address == IOAddress.REG_DMA2CNT_H:
            return self.dma_controller.channel_2.get_control_reg()
        elif address == IOAddress.REG_DMA3CNT_H:
            return self.dma_controller.channel_3.get_control_reg()
        elif address == IOAddress.REG_TM0CNT_L:
            return self.timers.timer_0.get_counter()
        elif address == IOAddress.REG_TM0CNT_H:
            return self.timers.timer_0.get_control_reg()
        elif address == IOAddress.REG_TM1CNT_L:
            return self.timers.timer_1.get_counter()
        elif address == IOAddress.REG_TM1CNT_H:
            return self.timers.timer_1.get_control_reg()
        elif address == IOAddress.REG_TM2CNT_L:
            return self.timers.timer_2.get_counter()
        elif address == IOAddress.REG_TM2CNT_H:
            return self.timers.timer_2.get_control_reg()
        elif address == IOAddress.REG_TM3CNT_L:
            return self.timers.timer_3.get_counter()
        elif address == IOAddress.REG_TM3CNT_H:
            return self.timers.timer_3.get_control_reg()
        elif address == IOAddress.REG_KEYINPUT:
            return self.keypad.key_input
        elif address == IOAddress.REG_KEYCNT:
            return self.keypad.key_control.reg
        elif address == IOAddress.REG_IE:
            return self.interrupt_controller.get_interrupt_enable()
        elif address == IOAddress.REG_IF:
            return self.interrupt_controller.get_interrupt_flags()
        elif address == IOAddress.REG_WAITCNT:
            return self.memory.wait_control.reg
        elif address == IOAddress.REG_IME:
            return self.interrupt_controller.get_interrupt_master_enable()
        else:
            return 0

    def read_8(self, address: int) -> int:
        value = self.read_16(address & ~1)
        return value >> (address & 1) * 8 & 255

    def write_32(self, address: int, value: int):
        self.write_16(address, value & 65535)
        self.write_16(address + 2, value >> 16 & 65535)

    def write_16(self, address: int, value: int):
        if address == IOAddress.REG_DISPCNT:
            self.ppu.display_control.reg = value
        elif address == IOAddress.REG_DISPSTAT:
            self.ppu.display_status.reg = value
        elif address == IOAddress.REG_VCOUNT:
            self.ppu.vcount = value
        elif address == IOAddress.REG_BG0CNT:
            self.ppu.bg_control_0.reg = value
        elif address == IOAddress.REG_BG1CNT:
            self.ppu.bg_control_1.reg = value
        elif address == IOAddress.REG_BG2CNT:
            self.ppu.bg_control_2.reg = value
        elif address == IOAddress.REG_BG3CNT:
            self.ppu.bg_control_3.reg = value
        elif address == IOAddress.REG_BG0HOFS:
            self.ppu.bg_offset_h[0] = value & 1023
        elif address == IOAddress.REG_BG0VOFS:
            self.ppu.bg_offset_v[0] = value & 1023
        elif address == IOAddress.REG_BG1HOFS:
            self.ppu.bg_offset_h[1] = value & 1023
        elif address == IOAddress.REG_BG1VOFS:
            self.ppu.bg_offset_v[1] = value & 1023
        elif address == IOAddress.REG_BG2HOFS:
            self.ppu.bg_offset_h[2] = value & 1023
        elif address == IOAddress.REG_BG2VOFS:
            self.ppu.bg_offset_v[2] = value & 1023
        elif address == IOAddress.REG_BG3HOFS:
            self.ppu.bg_offset_h[3] = value & 1023
        elif address == IOAddress.REG_BG3VOFS:
            self.ppu.bg_offset_v[3] = value & 1023
        elif address == IOAddress.REG_WIN0H:
            self.ppu.window_h_min[0] = value >> 8 & 255
            self.ppu.window_h_max[0] = value & 255
        elif address == IOAddress.REG_WIN1H:
            self.ppu.window_h_min[1] = value >> 8 & 255
            self.ppu.window_h_max[1] = value & 255
        elif address == IOAddress.REG_WIN0V:
            self.ppu.window_v_min[0] = value >> 8 & 255
            self.ppu.window_v_max[0] = value & 255
        elif address == IOAddress.REG_WIN1V:
            self.ppu.window_v_min[1] = value >> 8 & 255
            self.ppu.window_v_max[1] = value & 255
        elif address == IOAddress.REG_WININ:
            self.ppu.window_control_0.reg = value & 63
            self.ppu.window_control_1.reg = value >> 8 & 63
        elif address == IOAddress.REG_WINOUT:
            self.ppu.window_control_out.reg = value & 63
            self.ppu.window_control_obj.reg = value >> 8 & 63
        elif address == IOAddress.REG_BLDCNT:
            self.ppu.blend_control.reg = value
        elif address == IOAddress.REG_BLDALPHA:
            self.ppu.blend_alpha.reg = value
        elif address == IOAddress.REG_BLDY:
            self.ppu.blend_brightness.reg = value
        elif address == IOAddress.REG_SOUNDBIAS:
            self.reg_soundbias = value
        elif address == IOAddress.REG_DMA0SAD_L:
            mask = 4294901760
            self.dma_controller.channel_0.set_src_address(self.dma_controller.channel_0.get_src_address() & mask)
            self.dma_controller.channel_0.set_src_address(self.dma_controller.channel_0.get_src_address() | value)
        elif address == IOAddress.REG_DMA0SAD_H:
            mask = 65535
            self.dma_controller.channel_0.set_src_address(self.dma_controller.channel_0.get_src_address() & mask)
            self.dma_controller.channel_0.set_src_address(self.dma_controller.channel_0.get_src_address() | value << 16)
        elif address == IOAddress.REG_DMA0DAD_L:
            mask = 4294901760
            self.dma_controller.channel_0.set_dst_address(self.dma_controller.channel_0.get_dst_address() & mask)
            self.dma_controller.channel_0.set_dst_address(self.dma_controller.channel_0.get_dst_address() | value)
        elif address == IOAddress.REG_DMA0DAD_H:
            mask = 65535
            self.dma_controller.channel_0.set_dst_address(self.dma_controller.channel_0.get_dst_address() & mask)
            self.dma_controller.channel_0.set_dst_address(self.dma_controller.channel_0.get_dst_address() | value << 16)
        elif address == IOAddress.REG_DMA1SAD_L:
            mask = 4294901760
            self.dma_controller.channel_1.set_src_address(self.dma_controller.channel_1.get_src_address() & mask)
            self.dma_controller.channel_1.set_src_address(self.dma_controller.channel_1.get_src_address() | value)
        elif address == IOAddress.REG_DMA1SAD_H:
            mask = 65535
            self.dma_controller.channel_1.set_src_address(self.dma_controller.channel_1.get_src_address() & mask)
            self.dma_controller.channel_1.set_src_address(self.dma_controller.channel_1.get_src_address() | value << 16)
        elif address == IOAddress.REG_DMA1DAD_L:
            mask = 4294901760
            self.dma_controller.channel_1.set_dst_address(self.dma_controller.channel_1.get_dst_address() & mask)
            self.dma_controller.channel_1.set_dst_address(self.dma_controller.channel_1.get_dst_address() | value)
        elif address == IOAddress.REG_DMA1DAD_H:
            mask = 65535
            self.dma_controller.channel_1.set_dst_address(self.dma_controller.channel_1.get_dst_address() & mask)
            self.dma_controller.channel_1.set_dst_address(self.dma_controller.channel_1.get_dst_address() | value << 16)
        elif address == IOAddress.REG_DMA2SAD_L:
            mask = 4294901760
            self.dma_controller.channel_2.set_src_address(self.dma_controller.channel_2.get_src_address() & mask)
            self.dma_controller.channel_2.set_src_address(self.dma_controller.channel_2.get_src_address() | value)
        elif address == IOAddress.REG_DMA2SAD_H:
            mask = 65535
            self.dma_controller.channel_2.set_src_address(self.dma_controller.channel_2.get_src_address() & mask)
            self.dma_controller.channel_2.set_src_address(self.dma_controller.channel_2.get_src_address() | value << 16)
        elif address == IOAddress.REG_DMA2DAD_L:
            mask = 4294901760
            self.dma_controller.channel_2.set_dst_address(self.dma_controller.channel_2.get_dst_address() & mask)
            self.dma_controller.channel_2.set_dst_address(self.dma_controller.channel_2.get_dst_address() | value)
        elif address == IOAddress.REG_DMA2DAD_H:
            mask = 65535
            self.dma_controller.channel_2.set_dst_address(self.dma_controller.channel_2.get_dst_address() & mask)
            self.dma_controller.channel_2.set_dst_address(self.dma_controller.channel_2.get_dst_address() | value << 16)
        elif address == IOAddress.REG_DMA3SAD_L:
            mask = 4294901760
            self.dma_controller.channel_3.set_src_address(self.dma_controller.channel_3.get_src_address() & mask)
            self.dma_controller.channel_3.set_src_address(self.dma_controller.channel_3.get_src_address() | value)
        elif address == IOAddress.REG_DMA3SAD_H:
            mask = 65535
            self.dma_controller.channel_3.set_src_address(self.dma_controller.channel_3.get_src_address() & mask)
            self.dma_controller.channel_3.set_src_address(self.dma_controller.channel_3.get_src_address() | value << 16)
        elif address == IOAddress.REG_DMA3DAD_L:
            mask = 4294901760
            self.dma_controller.channel_3.set_dst_address(self.dma_controller.channel_3.get_dst_address() & mask)
            self.dma_controller.channel_3.set_dst_address(self.dma_controller.channel_3.get_dst_address() | value)
        elif address == IOAddress.REG_DMA3DAD_H:
            mask = 65535
            self.dma_controller.channel_3.set_dst_address(self.dma_controller.channel_3.get_dst_address() & mask)
            self.dma_controller.channel_3.set_dst_address(self.dma_controller.channel_3.get_dst_address() | value << 16)
        elif address == IOAddress.REG_DMA0CNT_L:
            self.dma_controller.channel_0.set_count(value)
        elif address == IOAddress.REG_DMA0CNT_H:
            self.dma_controller.channel_0.set_control_reg(value)
        elif address == IOAddress.REG_DMA1CNT_L:
            self.dma_controller.channel_1.set_count(value)
        elif address == IOAddress.REG_DMA1CNT_H:
            self.dma_controller.channel_1.set_control_reg(value)
        elif address == IOAddress.REG_DMA2CNT_L:
            self.dma_controller.channel_2.set_count(value)
        elif address == IOAddress.REG_DMA2CNT_H:
            self.dma_controller.channel_2.set_control_reg(value)
        elif address == IOAddress.REG_DMA3CNT_L:
            self.dma_controller.channel_3.set_count(value)
        elif address == IOAddress.REG_DMA3CNT_H:
            self.dma_controller.channel_3.set_control_reg(value)
        elif address == IOAddress.REG_TM0CNT_L:
            self.timers.timer_0.set_counter(value)
        elif address == IOAddress.REG_TM0CNT_H:
            self.timers.timer_0.set_control_reg(value)
        elif address == IOAddress.REG_TM1CNT_L:
            self.timers.timer_1.set_counter(value)
        elif address == IOAddress.REG_TM1CNT_H:
            self.timers.timer_1.set_control_reg(value)
        elif address == IOAddress.REG_TM2CNT_L:
            self.timers.timer_2.set_counter(value)
        elif address == IOAddress.REG_TM2CNT_H:
            self.timers.timer_2.set_control_reg(value)
        elif address == IOAddress.REG_TM3CNT_L:
            self.timers.timer_3.set_counter(value)
        elif address == IOAddress.REG_TM3CNT_H:
            self.timers.timer_3.set_control_reg(value)
        elif address == IOAddress.REG_KEYCNT:
            self.keypad.key_control.reg = value
        elif address == IOAddress.REG_IE:
            self.interrupt_controller.set_interrupt_enable(value)
        elif address == IOAddress.REG_IF:
            self.interrupt_controller.set_interrupt_flags(value)
        elif address == IOAddress.REG_WAITCNT:
            self.memory.wait_control.reg = value
            self.memory.update_waitstates()
        elif address == IOAddress.REG_IME:
            self.interrupt_controller.set_interrupt_master_enable(value)
        elif address == IOAddress.REG_HALTCNT:
            self.interrupt_controller.power_down_mode = get_bit(value, 15) + 1

    def write_8(self, address: int, value: int):
        aligned_address = address & ~1
        old_value = self.read_16(aligned_address)
        new_value = old_value & 255 | value << 8 if get_bit(address, 0) else old_value & 65280 | value
        self.write_16(aligned_address, new_value)


# ifndef CYTHON
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pyboy_advance.memory.memory import Memory

from pyboy_advance.constants import Interrupt, EventTrigger
from pyboy_advance.memory.constants import (
    IOAddress,
    MemoryAccess,
    DMAAddressAdjustment,
    DMAStartTiming,
    DMATransferSize,
)
from pyboy_advance.scheduler import Scheduler
from pyboy_advance.utils import get_bits, get_bit, set_bit, bint
# endif


class DMAControlRegister:
    def __init__(self):
        self.reg = 0

    @property
    def dst_address_adjustment(self) -> DMAAddressAdjustment | int:
        return get_bits(self.reg, 5, 6)

    @property
    def src_address_adjustment(self) -> DMAAddressAdjustment | int:
        return get_bits(self.reg, 7, 8)

    @property
    def repeat(self) -> bint:
        return get_bit(self.reg, 9)

    @property
    def transfer_size(self) -> DMATransferSize | int:
        return get_bit(self.reg, 10)

    @property
    def start_timing(self) -> DMAStartTiming | int:
        return get_bits(self.reg, 12, 13)

    @property
    def irq_when_done(self) -> bint:
        return get_bit(self.reg, 14)

    @property
    def transfer_enabled(self) -> bint:
        return get_bit(self.reg, 15)

    @transfer_enabled.setter
    def transfer_enabled(self, enable: bint):
        self.reg = set_bit(self.reg, 15, enable)


class DMAChannel:
    def __init__(self, channel_id: int, scheduler: Scheduler, memory: Memory):
        self.SRC_MASK = [0x07FFFFFF, 0x0FFFFFFF, 0x0FFFFFFF, 0x0FFFFFFF]
        self.DST_MASK = [0x07FFFFFF, 0x07FFFFFF, 0x07FFFFFF, 0x0FFFFFFF]
        self.COUNT_MASK = [0x3FFF, 0x3FFF, 0x3FFF, 0xFFFF]
        self.INTERRUPT = [Interrupt.DMA_0, Interrupt.DMA_1, Interrupt.DMA_2, Interrupt.DMA_3]
        self.TRANSFER_DELAY = 2

        self.channel_id = channel_id
        self.scheduler = scheduler
        self.memory = memory
        self.fifo = False
        self.pending = False

        self._control_reg = DMAControlRegister()
        self._count = 0
        self._src = 0
        self._dst = 0

        self._internal_count = 0
        self._internal_src = 0
        self._internal_dst = 0

        self._event = None

    @property
    def control_reg(self) -> int:
        return self._control_reg.reg

    @control_reg.setter
    def control_reg(self, value: int) -> None:
        old_enable = self._control_reg.transfer_enabled
        self._control_reg.reg = value

        if not old_enable and self._control_reg.transfer_enabled:  # DMA enabled
            self.fifo = (
                self._control_reg.start_timing == DMAStartTiming.SPECIAL
                and (self.channel_id == 1 or self.channel_id == 2)
                and (self._dst == IOAddress.REG_FIFO_A or self._dst == IOAddress.REG_FIFO_B)
            )

            self._internal_src = self._src
            self._internal_dst = self._dst
            self._internal_count = self._count

            # Count of 0 is treated as max count
            if self._internal_count == 0:
                self._internal_count = self.COUNT_MASK[self.channel_id] + 1

            if self._control_reg.start_timing == DMAStartTiming.IMMEDIATELY:
                self._event = self.scheduler.schedule(
                    self.activate, self.TRANSFER_DELAY, EventTrigger.IMMEDIATELY
                )
            elif self._control_reg.start_timing == DMAStartTiming.VBLANK:
                self._event = self.scheduler.schedule(
                    self.activate, self.TRANSFER_DELAY, EventTrigger.VBLANK
                )
            elif self._control_reg.start_timing == DMAStartTiming.HBLANK:
                self._event = self.scheduler.schedule(
                    self.activate, self.TRANSFER_DELAY, EventTrigger.HBLANK
                )

        elif old_enable and not self._control_reg.transfer_enabled:  # DMA cancelled
            if self._event:
                self._event.cancelled = False
                self._event = None
            self.pending = False

    @property
    def count(self) -> int:
        return self._count

    @count.setter
    def count(self, value: int) -> None:
        self._count = value & self.COUNT_MASK[self.channel_id]

    @property
    def src_address(self) -> int:
        return self._src

    @src_address.setter
    def src_address(self, value: int):
        self._src = value & self.SRC_MASK[self.channel_id]

    @property
    def dst_address(self) -> int:
        return self._dst

    @dst_address.setter
    def dst_address(self, value: int):
        self._dst = value & self.DST_MASK[self.channel_id]

    def transfer(self):
        if not self.pending:
            return

        transfer_size = self._control_reg.transfer_size
        transfer_size_bytes = 4 if transfer_size == DMATransferSize.WORD else 2

        address_alignment = ~0b11 if transfer_size == DMATransferSize.WORD else ~0b1
        self._internal_src &= address_alignment
        self._internal_dst &= address_alignment

        src_adj = self._control_reg.src_address_adjustment
        if src_adj == DMAAddressAdjustment.INCREMENT:
            src_step = transfer_size_bytes
        elif src_adj == DMAAddressAdjustment.DECREMENT:
            src_step = -transfer_size_bytes
        elif src_adj == DMAAddressAdjustment.LEAVE_UNCHANGED:
            src_step = 0
        # ifndef CYTHON
        else:
            assert False, "Unreachable, src_adj must be one of DMAAddressAdjustment"
        # endif

        dst_adj = self._control_reg.dst_address_adjustment
        if self.fifo:
            dst_step = 0
        elif dst_adj == DMAAddressAdjustment.INCREMENT:
            dst_step = transfer_size_bytes
        elif dst_adj == DMAAddressAdjustment.DECREMENT:
            dst_step = -transfer_size_bytes
        elif dst_adj == DMAAddressAdjustment.LEAVE_UNCHANGED:
            dst_step = 0
        elif dst_adj == DMAAddressAdjustment.INCREMENT_RELOAD:
            dst_step = transfer_size_bytes
        # ifndef CYTHON
        else:
            assert False, "Unreachable, dst_adj must be one of DMAAddressAdjustment"
        # endif

        access = MemoryAccess.NON_SEQUENTIAL

        for _ in range(self._internal_count):
            if transfer_size == DMATransferSize.WORD:
                value = self.memory.read_32(self._internal_src, access)
                self.memory.write_32(self._internal_dst, value, access)
            else:
                value = self.memory.read_16(self._internal_src, access)
                self.memory.write_16(self._internal_dst, value, access)

            self._internal_src += src_step
            self._internal_dst += dst_step
            access = MemoryAccess.SEQUENTIAL

        if self._control_reg.irq_when_done:
            self.memory.io.interrupt_controller.signal(self.INTERRUPT[self.channel_id])

        self.pending = False

        if self._control_reg.repeat:
            if dst_adj == DMAAddressAdjustment.INCREMENT_RELOAD:
                self._internal_dst = self._dst

            if self._control_reg.start_timing == DMAStartTiming.VBLANK:
                self._event = self.scheduler.schedule(
                    self.activate, self.TRANSFER_DELAY, EventTrigger.VBLANK
                )
            elif self._control_reg.start_timing == DMAStartTiming.HBLANK:
                self._event = self.scheduler.schedule(
                    self.activate, self.TRANSFER_DELAY, EventTrigger.HBLANK
                )
        else:
            self._control_reg.transfer_enabled = False

    def activate(self):
        if self._control_reg.transfer_enabled:
            self.pending = True


class DMAController:
    def __init__(self, scheduler: Scheduler, memory: Memory):
        self.channel_0 = DMAChannel(0, scheduler, memory)
        self.channel_1 = DMAChannel(1, scheduler, memory)
        self.channel_2 = DMAChannel(2, scheduler, memory)
        self.channel_3 = DMAChannel(3, scheduler, memory)

    @property
    def active(self) -> bool:
        return (
            self.channel_0.pending
            or self.channel_1.pending
            or self.channel_2.pending
            or self.channel_3.pending
        )

    def perform_transfers(self):
        self.channel_0.transfer()
        self.channel_1.transfer()
        self.channel_2.transfer()
        self.channel_3.transfer()



class VideoMemory:

    def __init__(self, display_control: DisplayControlRegister):
        self.display_control = display_control
        self.palram = array('B', [0] * MemoryRegion.PALRAM_SIZE)
        self.vram = array('B', [0] * MemoryRegion.VRAM_SIZE)
        self.oam = array('B', [0] * MemoryRegion.OAM_SIZE)

    def read_32_palram(self, address: int) -> int:
        return array_read_32(self.palram, address & MemoryRegion.PALRAM_MASK)

    def read_32_vram(self, address: int) -> int:
        mask = self._get_vram_address_mask(address)
        return array_read_32(self.vram, address & mask)

    def read_32_oam(self, address: int) -> int:
        return array_read_32(self.oam, address & MemoryRegion.OAM_MASK)

    def read_16_palram(self, address: int) -> int:
        return array_read_16(self.palram, address & MemoryRegion.PALRAM_MASK)

    def read_16_vram(self, address: int) -> int:
        mask = self._get_vram_address_mask(address)
        return array_read_16(self.vram, address & mask)

    def read_16_oam(self, address: int) -> int:
        return array_read_16(self.oam, address & MemoryRegion.OAM_MASK)

    def read_8_palram(self, address: int) -> int:
        return self.palram[address & MemoryRegion.PALRAM_MASK]

    def read_8_vram(self, address: int) -> int:
        mask = self._get_vram_address_mask(address)
        return self.vram[address & mask]

    def read_8_oam(self, address: int) -> int:
        return self.oam[address & MemoryRegion.OAM_MASK]

    def write_32_palram(self, address: int, value: int):
        array_write_32(self.palram, address & MemoryRegion.PALRAM_MASK, value)

    def write_32_vram(self, address: int, value: int):
        mask = self._get_vram_address_mask(address)
        array_write_32(self.vram, address & mask, value)

    def write_32_oam(self, address: int, value: int):
        array_write_32(self.oam, address & MemoryRegion.OAM_MASK, value)

    def write_16_palram(self, address: int, value: int):
        array_write_16(self.palram, address & MemoryRegion.PALRAM_MASK, value)

    def write_16_vram(self, address: int, value: int):
        mask = self._get_vram_address_mask(address)
        array_write_16(self.vram, address & mask, value)

    def write_16_oam(self, address: int, value: int):
        array_write_16(self.oam, address & MemoryRegion.OAM_MASK, value)

    def write_8_palram(self, address: int, value: int):
        address = address & ~1
        self.palram[address & MemoryRegion.PALRAM_MASK] = value
        self.palram[address + 1 & MemoryRegion.PALRAM_MASK] = value

    def write_8_vram(self, address: int, value: int):
        video_mode = self.display_control.get_video_mode()
        bitmapped = video_mode in [VideoMode.MODE_3, VideoMode.MODE_4, VideoMode.MODE_5]
        bg_region_end = 81920 if bitmapped else 65536
        if address & 131071 < bg_region_end:
            address = address & ~1
            mask_1 = self._get_vram_address_mask(address)
            mask_2 = self._get_vram_address_mask(address + 1)
            self.vram[address & mask_1] = value
            self.vram[address + 1 & mask_2] = value

    def _get_vram_address_mask(self, address: int):
        """
        Get the appropriate mask for the given VRAM address.

        GBA VRAM is 96 KB (64K + 32K) but is mirrored in 128 KB steps
        (64K + 32K + 32K, the two 32K blocks being mirrors of each other).

        Hence, if the address is in the mirrored upper region (bit 16 set), we use
        VRAM_MASK_1 to wrap it back into valid VRAM space. Otherwise, we use VRAM_MASK_2.
        """
        return MemoryRegion.VRAM_MASK_1 if get_bit(address, 16) else MemoryRegion.VRAM_MASK_2


cimport cython

from libc.stdint cimport uint32_t

from pyboy_advance.cpu.cpu cimport CPU
from pyboy_advance.cpu.arm.alu cimport (
    arm_alu_sub,
    arm_alu_add,
    arm_alu_mov,
    arm_alu_adc,
    arm_alu_sbc,
    arm_alu_tst,
    arm_alu_cmp,
    arm_alu_cmn,
    arm_alu_orr,
    arm_alu_bic,
    arm_alu_mvn,
    arm_alu_and,
    arm_alu_eor,
)
from pyboy_advance.cpu.arm.mul cimport arm_multiply_idle
from pyboy_advance.cpu.constants cimport ShiftType
from pyboy_advance.cpu.thumb.constants cimport ThumbALUOpcode
from pyboy_advance.memory.constants cimport MemoryAccess
from pyboy_advance.utils cimport (
    get_bits,
    get_bit,
    sign_32,
    add_32,
)

cdef void thumb_move_shifted_register(CPU, uint32_t) noexcept

cdef void thumb_add_subtract(CPU, uint32_t) noexcept

cdef void thumb_move_compare_add_subtract(CPU, uint32_t) noexcept

cdef void thumb_alu(CPU, uint32_t) noexcept

cdef void thumb_alu_shift(CPU, uint32_t, uint32_t, uint32_t, int) noexcept

@cython.locals(mask=uint32_t)
cdef void thumb_alu_multiply(CPU, uint32_t, uint32_t, uint32_t) noexcept

cdef void thumb_high_reg_branch_exchange(CPU, uint32_t) noexcept

cdef void thumb_add_offset_to_stack_pointer(CPU, uint32_t) noexcept

cdef void thumb_get_address(CPU, uint32_t) noexcept



import pytest


@pytest.mark.rom_url(
    "https://raw.githubusercontent.com/nba-emu/hw-test/master/timer/reload/reload.gba"
)
def test_nba_timer_reload(emulator, assert_expected_image):
    emulator.frame(20)
    assert_expected_image(emulator.screen.image)


@pytest.mark.rom_url(
    "https://raw.githubusercontent.com/nba-emu/hw-test/master/timer/start-stop/start-stop.gba"
)
def test_nba_timer_start_stop(emulator, assert_expected_image):
    emulator.frame(20)
    assert_expected_image(emulator.screen.image)



# ifndef CYTHON
from __future__ import annotations

from array import array
from typing import TYPE_CHECKING, Iterable

if TYPE_CHECKING:
    from pyboy_advance.cpu.cpu import CPU

from pyboy_advance.cpu.constants import CPUState
from pyboy_advance.memory.constants import MemoryRegion, MemoryAccess
from pyboy_advance.memory.gamepak import GamePak
from pyboy_advance.memory.io import IO
from pyboy_advance.scheduler import Scheduler
from pyboy_advance.utils import (
    get_bit,
    array_read_16,
    array_read_32,
    array_write_32,
    array_write_16,
    ror_32,
    extend_sign_16,
    extend_sign_8,
    get_bits,
)
# endif


class Memory:
    def __init__(
        self,
        scheduler: Scheduler,
        gamepak: GamePak,
        bios: bytes | bytearray | Iterable[int],
    ):
        self.scheduler = scheduler
        self.cpu: CPU | None = None
        self.io: IO | None = None

        # General Internal Memory
        self.bios = array("B", bios.ljust(MemoryRegion.BIOS_SIZE))
        self.ewram = array("B", [0] * MemoryRegion.EWRAM_SIZE)
        self.iwram = array("B", [0] * MemoryRegion.IWRAM_SIZE)

        # External Memory (Game Pak)
        self.gamepak = gamepak

        # If the program counter is not in the BIOS region, reading
        # BIOS will return the most recently fetched BIOS opcode,
        # which is tracked by this variable
        self.bios_last_opcode = 0

        self.wait_control = WaitstateControlRegister()

        self._init_access_times()

    def read_32(self, address: int, access_type: MemoryAccess) -> int:
        return self._read_32_internal(address, access_type)

    def read_32_ror(self, address: int, access_type: MemoryAccess) -> int:
        value = self.read_32(address, access_type)
        rotate = (address & 0b11) * 8
        return ror_32(value, rotate)

    def read_16(self, address: int, access_type: MemoryAccess) -> int:
        return self._read_16_internal(address, access_type)

    def read_16_signed(self, address: int, access_type: MemoryAccess) -> int:
        if get_bit(address, 0):
            return self.read_8_signed(address, access_type)
        return extend_sign_16(self.read_16(address, access_type))

    def read_16_ror(self, address: int, access_type: MemoryAccess) -> int:
        value = self.read_16(address, access_type)
        rotate = (address & 0b1) * 8
        return ror_32(value, rotate)

    def read_8(self, address: int, access_type: MemoryAccess) -> int:
        return self._read_8_internal(address, access_type)

    def read_8_signed(self, address: int, access_type: MemoryAccess) -> int:
        value = self.read_8(address, access_type)
        return extend_sign_8(value)

    def write_32(self, address: int, value: int, access_type: MemoryAccess):
        self._write_32_internal(address, value, access_type)

    def write_16(self, address: int, value: int, access_type: MemoryAccess):
        self._write_16_internal(address, value, access_type)

    def write_8(self, address: int, value: int, access_type: MemoryAccess):
        self._write_8_internal(address, value, access_type)

    def _read_32_internal(self, address: int, access_type: MemoryAccess) -> int:
        self._idle_for_access(address, 4, access_type)

        address = address & ~0b11  # Align address to 4-byte boundary
        region = address >> 24

        if region == MemoryRegion.BIOS_REGION:
            if address <= MemoryRegion.BIOS_END:
                if self.cpu.regs.pc <= MemoryRegion.BIOS_END:
                    self.bios_last_opcode = array_read_32(
                        self.bios, address & MemoryRegion.BIOS_MASK
                    )
                return self.bios_last_opcode
            return self._read_unused_memory()

        elif region == MemoryRegion.EWRAM_REGION:
            return array_read_32(self.ewram, address & MemoryRegion.EWRAM_MASK)

        elif region == MemoryRegion.IWRAM_REGION:
            return array_read_32(self.iwram, address & MemoryRegion.IWRAM_MASK)

        elif MemoryRegion.IO_START <= address <= MemoryRegion.IO_END:
            return self.io.read_32(address)

        elif region == MemoryRegion.PALRAM_REGION:
            return self.io.ppu.memory.read_32_palram(address)

        elif region == MemoryRegion.VRAM_REGION:
            return self.io.ppu.memory.read_32_vram(address)

        elif region == MemoryRegion.OAM_REGION:
            return self.io.ppu.memory.read_32_oam(address)

        elif MemoryRegion.GAMEPAK_REGION_START <= region <= MemoryRegion.GAMEPAK_REGION_END:
            return self.gamepak.read_32(address)

        elif region == MemoryRegion.SRAM_REGION:
            print(f"Attempt to read SRAM: {address:#010x}")
            return 0

        else:
            return self._read_unused_memory()

    def _read_16_internal(self, address: int, access_type: MemoryAccess) -> int:
        self._idle_for_access(address, 2, access_type)

        address = address & ~0b1  # Align address to 2-byte boundary
        region = address >> 24

        if region == MemoryRegion.BIOS_REGION:
            if address <= MemoryRegion.BIOS_END:
                if self.cpu.regs.pc <= MemoryRegion.BIOS_END:
                    self.bios_last_opcode = array_read_32(
                        self.bios, address & ~0b11 & MemoryRegion.BIOS_MASK
                    )
                return (self.bios_last_opcode >> ((address & 0b11) * 8)) & 0xFFFF
            return (self._read_unused_memory() >> ((address & 0b11) * 8)) & 0xFFFF

        elif region == MemoryRegion.EWRAM_REGION:
            return array_read_16(self.ewram, address & MemoryRegion.EWRAM_MASK)

        elif region == MemoryRegion.IWRAM_REGION:
            return array_read_16(self.iwram, address & MemoryRegion.IWRAM_MASK)

        elif MemoryRegion.IO_START <= address <= MemoryRegion.IO_END:
            return self.io.read_16(address)

        elif region == MemoryRegion.PALRAM_REGION:
            return self.io.ppu.memory.read_16_palram(address)

        elif region == MemoryRegion.VRAM_REGION:
            return self.io.ppu.memory.read_16_vram(address)

        elif region == MemoryRegion.OAM_REGION:
            return self.io.ppu.memory.read_16_oam(address)

        elif MemoryRegion.GAMEPAK_REGION_START <= region <= MemoryRegion.GAMEPAK_REGION_END:
            return self.gamepak.read_16(address)

        elif region == MemoryRegion.SRAM_REGION:
            print(f"Attempt to read SRAM: {address:#010x}")
            return 0

        else:
            return (self._read_unused_memory() >> ((address & 0b11) * 8)) & 0xFFFF

    def _read_8_internal(self, address: int, access_type: MemoryAccess) -> int:
        self._idle_for_access(address, 1, access_type)

        region = address >> 24

        if region == MemoryRegion.BIOS_REGION:
            if address <= MemoryRegion.BIOS_END:
                if self.cpu.regs.pc <= MemoryRegion.BIOS_END:
                    self.bios_last_opcode = array_read_32(
                        self.bios, address & ~0b11 & MemoryRegion.BIOS_MASK
                    )
                return (self.bios_last_opcode >> ((address & 0b11) * 8)) & 0xFF
            return (self._read_unused_memory() >> ((address & 0b11) * 8)) & 0xFF

        elif region == MemoryRegion.EWRAM_REGION:
            return self.ewram[address & MemoryRegion.EWRAM_MASK]

        elif region == MemoryRegion.IWRAM_REGION:
            return self.iwram[address & MemoryRegion.IWRAM_MASK]

        elif MemoryRegion.IO_START <= address <= MemoryRegion.IO_END:
            return self.io.read_8(address)

        elif region == MemoryRegion.PALRAM_REGION:
            return self.io.ppu.memory.read_8_palram(address)

        elif region == MemoryRegion.VRAM_REGION:
            return self.io.ppu.memory.read_8_vram(address)

        elif region == MemoryRegion.OAM_REGION:
            return self.io.ppu.memory.read_8_oam(address)

        elif MemoryRegion.GAMEPAK_REGION_START <= region <= MemoryRegion.GAMEPAK_REGION_END:
            return self.gamepak.read_8(address)

        elif region == MemoryRegion.SRAM_REGION:
            print(f"Attempt to read SRAM: {address:#010x}")
            return 0

        else:
            return (self._read_unused_memory() >> ((address & 0b11) * 8)) & 0xFF

    def _write_32_internal(self, address: int, value: int, access_type: MemoryAccess):
        self._idle_for_access(address, 4, access_type)

        address = address & ~0b11  # Align address to 4-byte boundary
        region = address >> 24

        if region == MemoryRegion.BIOS_REGION:
            # Ignore attempts to write to BIOS region
            pass

        elif region == MemoryRegion.EWRAM_REGION:
            array_write_32(self.ewram, address & MemoryRegion.EWRAM_MASK, value)

        elif region == MemoryRegion.IWRAM_REGION:
            array_write_32(self.iwram, address & MemoryRegion.IWRAM_MASK, value)

        elif region == MemoryRegion.IO_REGION:
            self.io.write_32(address, value)

        elif region == MemoryRegion.PALRAM_REGION:
            self.io.ppu.memory.write_32_palram(address, value)

        elif region == MemoryRegion.VRAM_REGION:
            self.io.ppu.memory.write_32_vram(address, value)

        elif region == MemoryRegion.OAM_REGION:
            self.io.ppu.memory.write_32_oam(address, value)

        elif MemoryRegion.GAMEPAK_REGION_START <= region <= MemoryRegion.GAMEPAK_REGION_END:
            print(f"Attempt to write to SRAM: {address:#010x}")

        elif region == MemoryRegion.SRAM_REGION:
            print(f"Attempt to write to SRAM: {address:#010x}")

        else:
            print(f"Attempt to write to unused memory: {address:#010x}")

    def _write_16_internal(self, address: int, value: int, access_type: MemoryAccess):
        self._idle_for_access(address, 2, access_type)

        address = address & ~0b1  # Align address to 2-byte boundary
        value = value & 0xFFFF
        region = address >> 24

        if region == MemoryRegion.BIOS_REGION:
            # Ignore attempts to write to BIOS region
            pass

        elif region == MemoryRegion.EWRAM_REGION:
            array_write_16(self.ewram, address & MemoryRegion.EWRAM_MASK, value)

        elif region == MemoryRegion.IWRAM_REGION:
            array_write_16(self.iwram, address & MemoryRegion.IWRAM_MASK, value)

        elif region == MemoryRegion.IO_REGION:
            self.io.write_16(address, value)

        elif region == MemoryRegion.PALRAM_REGION:
            self.io.ppu.memory.write_16_palram(address, value)

        elif region == MemoryRegion.VRAM_REGION:
            self.io.ppu.memory.write_16_vram(address, value)

        elif region == MemoryRegion.OAM_REGION:
            self.io.ppu.memory.write_16_oam(address, value)

        elif MemoryRegion.GAMEPAK_REGION_START <= region <= MemoryRegion.GAMEPAK_REGION_END:
            print(f"Attempt to write to SRAM: {address:#010x}")

        elif region == MemoryRegion.SRAM_REGION:
            print(f"Attempt to write to SRAM: {address:#010x}")

        else:
            print(f"Attempt to write to unused memory: {address:#010x}")

    def _write_8_internal(self, address: int, value: int, access_type: MemoryAccess):
        self._idle_for_access(address, 1, access_type)

        value = value & 0xFF
        region = address >> 24

        if region == MemoryRegion.BIOS_REGION:
            # Ignore attempts to write to BIOS region
            pass

        elif region == MemoryRegion.EWRAM_REGION:
            self.ewram[address & MemoryRegion.EWRAM_MASK] = value

        elif region == MemoryRegion.IWRAM_REGION:
            self.iwram[address & MemoryRegion.IWRAM_MASK] = value

        elif region == MemoryRegion.IO_REGION:
            self.io.write_8(address, value)

        elif region == MemoryRegion.PALRAM_REGION:
            self.io.ppu.memory.write_8_palram(address, value)

        elif region == MemoryRegion.VRAM_REGION:
            self.io.ppu.memory.write_8_vram(address, value)

        elif region == MemoryRegion.OAM_REGION:
            # Ignore attempts to write a byte into OAM
            pass

        elif MemoryRegion.GAMEPAK_REGION_START <= region <= MemoryRegion.GAMEPAK_REGION_END:
            print(f"Attempt to write to SRAM: {address:#010x}")

        elif region == MemoryRegion.SRAM_REGION:
            print(f"Attempt to write to SRAM: {address:#010x}")

        else:
            print(f"Attempt to write to unused memory: {address:#010x}")

    def _read_unused_memory(self) -> int:
        """
        Reading from unused memory returns the last prefetched instruction.
        https://problemkaputt.de/gbatek.htm#gbaunpredictablethings

        For THUMB code, the result consists of two 16-bit fragments and depends on
        the address area and alignment where the opcode was stored.
        """
        if self.cpu.regs.cpsr.state == CPUState.ARM:
            return self.cpu.pipeline[1]

        region = self.cpu.regs.pc >> 24

        if region == MemoryRegion.BIOS_REGION or region == MemoryRegion.OAM_REGION:
            if (self.cpu.regs.pc & 0b11) == 0:
                # According to GBATEK, we should be using [$+6] for the top 16 bits
                # here, but that isn't prefetched yet?
                return (self.cpu.pipeline[1] << 16) | self.cpu.pipeline[1]
            else:
                return (self.cpu.pipeline[1] << 16) | self.cpu.pipeline[0]

        elif region == MemoryRegion.IWRAM_REGION:
            if (self.cpu.regs.pc & 0b11) == 0:
                return (self.cpu.pipeline[0] << 16) | self.cpu.pipeline[1]
            else:
                return (self.cpu.pipeline[1] << 16) | self.cpu.pipeline[0]

        else:
            return (self.cpu.pipeline[1] << 16) | self.cpu.pipeline[1]

    def _init_access_times(self):
        """
        Source: GBATEK

        Region        Bus   Read      Write     Cycles
        =====================================================
        BIOS ROM      32    8/16/32   -         1/1/1
        Work RAM 32K  32    8/16/32   8/16/32   1/1/1
        I/O           32    8/16/32   8/16/32   1/1/1
        OAM           32    8/16/32   16/32     1/1/1 *
        Work RAM 256K 16    8/16/32   8/16/32   3/3/6 **
        Palette RAM   16    8/16/32   16/32     1/1/2 *
        VRAM          16    8/16/32   16/32     1/1/2 *
        GamePak ROM   16    8/16/32   -         5/5/8 **/***
        GamePak Flash 16    8/16/32   16/32     5/5/8 **/***
        GamePak SRAM  8     8         8         5     **

        Timing Notes:

          *   Plus 1 cycle if GBA accesses video memory at the same time.
          **  Default waitstate settings, see System Control chapter.
          *** Separate timings for sequential, and non-sequential accesses.

        """
        # fmt: off
        self.access_time_32 = [[1] * 16 for _ in range(len(MemoryAccess))]
        self.access_time_16 = [[1] * 16 for _ in range(len(MemoryAccess))]

        self.access_time_32[int(MemoryAccess.NON_SEQUENTIAL)][int(MemoryRegion.EWRAM_REGION)]  = 6
        self.access_time_32[int(MemoryAccess.SEQUENTIAL)][int(MemoryRegion.EWRAM_REGION)]      = 6
        self.access_time_16[int(MemoryAccess.NON_SEQUENTIAL)][int(MemoryRegion.EWRAM_REGION)]  = 3
        self.access_time_16[int(MemoryAccess.SEQUENTIAL)][int(MemoryRegion.EWRAM_REGION)]      = 3

        self.access_time_32[int(MemoryAccess.NON_SEQUENTIAL)][int(MemoryRegion.PALRAM_REGION)] = 2
        self.access_time_32[int(MemoryAccess.SEQUENTIAL)][int(MemoryRegion.PALRAM_REGION)]     = 2

        self.access_time_32[int(MemoryAccess.NON_SEQUENTIAL)][int(MemoryRegion.VRAM_REGION)]   = 2
        self.access_time_32[int(MemoryAccess.SEQUENTIAL)][int(MemoryRegion.VRAM_REGION)]       = 2

        self.access_time_32[int(MemoryAccess.NON_SEQUENTIAL)][int(MemoryRegion.OAM_REGION)]    = 2
        self.access_time_32[int(MemoryAccess.SEQUENTIAL)][int(MemoryRegion.OAM_REGION)]        = 2

        self.update_waitstates()
        # fmt: on

    def update_waitstates(self):
        # fmt: off
        self.access_time_16[int(MemoryAccess.NON_SEQUENTIAL)][int(MemoryRegion.GAMEPAK_0_REGION_1)] = 1 + self.wait_control.ws0_non_seq
        self.access_time_16[int(MemoryAccess.NON_SEQUENTIAL)][int(MemoryRegion.GAMEPAK_0_REGION_2)] = 1 + self.wait_control.ws0_non_seq
        self.access_time_16[int(MemoryAccess.NON_SEQUENTIAL)][int(MemoryRegion.GAMEPAK_1_REGION_1)] = 1 + self.wait_control.ws1_non_seq
        self.access_time_16[int(MemoryAccess.NON_SEQUENTIAL)][int(MemoryRegion.GAMEPAK_1_REGION_2)] = 1 + self.wait_control.ws1_non_seq
        self.access_time_16[int(MemoryAccess.NON_SEQUENTIAL)][int(MemoryRegion.GAMEPAK_2_REGION_1)] = 1 + self.wait_control.ws2_non_seq
        self.access_time_16[int(MemoryAccess.NON_SEQUENTIAL)][int(MemoryRegion.GAMEPAK_2_REGION_2)] = 1 + self.wait_control.ws2_non_seq
        self.access_time_16[int(MemoryAccess.NON_SEQUENTIAL)][int(MemoryRegion.SRAM_REGION)]        = 1 + self.wait_control.sram

        self.access_time_16[int(MemoryAccess.SEQUENTIAL)][int(MemoryRegion.GAMEPAK_0_REGION_1)] = 1 + self.wait_control.ws0_seq
        self.access_time_16[int(MemoryAccess.SEQUENTIAL)][int(MemoryRegion.GAMEPAK_0_REGION_2)] = 1 + self.wait_control.ws0_seq
        self.access_time_16[int(MemoryAccess.SEQUENTIAL)][int(MemoryRegion.GAMEPAK_1_REGION_1)] = 1 + self.wait_control.ws1_seq
        self.access_time_16[int(MemoryAccess.SEQUENTIAL)][int(MemoryRegion.GAMEPAK_1_REGION_2)] = 1 + self.wait_control.ws1_seq
        self.access_time_16[int(MemoryAccess.SEQUENTIAL)][int(MemoryRegion.GAMEPAK_2_REGION_1)] = 1 + self.wait_control.ws2_seq
        self.access_time_16[int(MemoryAccess.SEQUENTIAL)][int(MemoryRegion.GAMEPAK_2_REGION_2)] = 1 + self.wait_control.ws2_seq
        self.access_time_16[int(MemoryAccess.SEQUENTIAL)][int(MemoryRegion.SRAM_REGION)]        = 1 + self.wait_control.sram
        # fmt: on

        for region in range(MemoryRegion.GAMEPAK_0_REGION_1, MemoryRegion.SRAM_REGION + 1):
            self.access_time_32[int(MemoryAccess.NON_SEQUENTIAL)][int(region)] = (
                self.access_time_16[int(MemoryAccess.NON_SEQUENTIAL)][int(region)]
                + self.access_time_16[int(MemoryAccess.SEQUENTIAL)][int(region)]
            )
            self.access_time_32[int(MemoryAccess.SEQUENTIAL)][int(region)] = (
                self.access_time_16[int(MemoryAccess.SEQUENTIAL)][int(region)] * 2
            )

    def _idle_for_access(self, address: int, size: int, access_type: MemoryAccess):
        region = (address >> 24) & 0xF

        cycles = (
            self.access_time_32[int(access_type)][int(region)]
            if size == 4
            else self.access_time_16[int(access_type)][int(region)]
        )

        self.scheduler.idle(cycles)


class WaitstateControlRegister:
    def __init__(self):
        self.NON_SEQUENTIAL_CYCLES = [4, 3, 2, 8]
        self.reg = 0

    @property
    def sram(self):
        return self.NON_SEQUENTIAL_CYCLES[get_bits(self.reg, 0, 1)]

    @property
    def ws0_non_seq(self):
        return self.NON_SEQUENTIAL_CYCLES[get_bits(self.reg, 2, 3)]

    @property
    def ws0_seq(self):
        return 1 if get_bit(self.reg, 4) else 2

    @property
    def ws1_non_seq(self):
        return self.NON_SEQUENTIAL_CYCLES[get_bits(self.reg, 5, 6)]

    @property
    def ws1_seq(self):
        return 1 if get_bit(self.reg, 7) else 4

    @property
    def ws2_non_seq(self):
        return self.NON_SEQUENTIAL_CYCLES[get_bits(self.reg, 8, 9)]

    @property
    def ws2_seq(self):
        return 1 if get_bit(self.reg, 10) else 8



from libc.stdint cimport uint8_t, uint16_t, uint32_t
from cpython.array cimport array

from pyboy_advance.constants cimport Interrupt, EventTrigger
from pyboy_advance.interrupt_controller cimport InterruptController
from pyboy_advance.ppu.constants cimport *
from pyboy_advance.ppu.memory cimport VideoMemory
from pyboy_advance.ppu.registers cimport (
    DisplayControlRegister,
    DisplayStatusRegister,
    BackgroundControlRegister,
    WindowControlRegister,
    BlendControlRegister,
    BlendAlphaRegister,
    BlendBrightnessRegister,
)
from pyboy_advance.scheduler cimport Scheduler
from pyboy_advance.utils cimport (
    get_bit,
    get_bits,
    sign_32,
    add_32,
    extend_sign_9,
)

cdef class Object:
    cdef (uint32_t, uint32_t)[3][4] SIZES

    cdef uint32_t attr_0
    cdef uint32_t attr_1
    cdef uint32_t attr_2

    cdef uint32_t get_x(self) noexcept
    cdef uint32_t get_y(self) noexcept
    cdef bint get_affine(self) noexcept
    cdef bint get_double_size(self) noexcept
    cdef bint get_disabled(self) noexcept
    cdef int get_mode(self) noexcept
    cdef bint get_mosaic(self) noexcept
    cdef bint get_colour_256(self) noexcept
    cdef int get_shape(self) noexcept
    cdef bint get_flip_horizontal(self) noexcept
    cdef bint get_flip_vertical(self) noexcept
    cdef (uint32_t, uint32_t) get_size(self) noexcept
    cdef uint32_t get_tile_index(self) noexcept
    cdef uint32_t get_priority(self) noexcept
    cdef uint32_t get_palette_num(self) noexcept

cdef class PPU:
    cdef Scheduler scheduler
    cdef InterruptController interrupt_controller
    cdef bool rendering_enabled
    cdef DisplayControlRegister display_control
    cdef DisplayStatusRegister display_status
    cdef uint32_t vcount
    cdef BackgroundControlRegister bg_control_0
    cdef BackgroundControlRegister bg_control_1
    cdef BackgroundControlRegister bg_control_2
    cdef BackgroundControlRegister bg_control_3
    cdef uint32_t[:] bg_offset_h
    cdef uint32_t[:] bg_offset_v
    cdef uint32_t[:] window_v_min
    cdef uint32_t[:] window_v_max
    cdef uint32_t[:] window_h_min
    cdef uint32_t[:] window_h_max
    cdef WindowControlRegister window_control_0
    cdef WindowControlRegister window_control_1
    cdef WindowControlRegister window_control_obj
    cdef WindowControlRegister window_control_out
    cdef uint16_t[:] window_mask_0
    cdef uint16_t[:] window_mask_1
    cdef uint16_t[:] window_mask_obj
    cdef BlendControlRegister blend_control
    cdef BlendAlphaRegister blend_alpha
    cdef BlendBrightnessRegister blend_brightness
    cdef bint[4][240] blend_mask_obj
    cdef VideoMemory memory
    cdef uint16_t[4][240] bg_scanline
    cdef uint16_t[4][240] obj_scanline
    cdef uint16_t[240] scanline
    cdef uint16_t[240] scanline_top_colour
    cdef uint32_t[240] scanline_top_layer
    cdef object front_buffer
    cdef object back_buffer
    cdef object front_buffer_ptr
    cdef object back_buffer_ptr

    cdef void hblank_start(self) noexcept
    cdef void hblank_end(self) noexcept
    cdef void _init_layers(self) noexcept
    cdef void _render_backgrounds(self) noexcept
    cdef void _render_background_text(self, int) noexcept
    cdef void _render_background_affine(self, int) noexcept
    cdef void _render_background_bitmap(self, bint, bint) noexcept
    cdef void _render_objects(self) noexcept
    cdef void _render_object(self, Object) noexcept
    cdef void _merge_layers(self) noexcept
    cdef void _merge_layer(self, uint16_t[:], int, int) noexcept
    cdef void _calc_window_masks(self) noexcept
    cdef WindowControlRegister _get_window_for_pixel(self, uint32_t) noexcept
    cdef uint32_t _get_palette_index(self, uint32_t, uint32_t, uint32_t, uint32_t, bint, uint32_t) noexcept
    cdef BackgroundControlRegister _get_bg_control(self, int) noexcept

cdef (uint8_t, uint8_t, uint8_t) get_rgb_channels(uint16_t) noexcept
cdef uint16_t blend_colours(uint16_t, uint16_t, uint32_t, uint32_t) noexcept



# ifndef CYTHON
from array import array

from pyboy_advance.memory.constants import MemoryRegion
from pyboy_advance.ppu.constants import VideoMode
from pyboy_advance.ppu.registers import DisplayControlRegister
from pyboy_advance.utils import (
    get_bit,
    array_read_32,
    array_read_16,
    array_write_32,
    array_write_16,
)
# endif


class VideoMemory:
    def __init__(self, display_control: DisplayControlRegister):
        self.display_control = display_control

        self.palram = array("B", [0] * MemoryRegion.PALRAM_SIZE)
        self.vram = array("B", [0] * MemoryRegion.VRAM_SIZE)
        self.oam = array("B", [0] * MemoryRegion.OAM_SIZE)

    def read_32_palram(self, address: int) -> int:
        return array_read_32(self.palram, address & MemoryRegion.PALRAM_MASK)

    def read_32_vram(self, address: int) -> int:
        mask = self._get_vram_address_mask(address)
        return array_read_32(self.vram, address & mask)

    def read_32_oam(self, address: int) -> int:
        return array_read_32(self.oam, address & MemoryRegion.OAM_MASK)

    def read_16_palram(self, address: int) -> int:
        return array_read_16(self.palram, address & MemoryRegion.PALRAM_MASK)

    def read_16_vram(self, address: int) -> int:
        mask = self._get_vram_address_mask(address)
        return array_read_16(self.vram, address & mask)

    def read_16_oam(self, address: int) -> int:
        return array_read_16(self.oam, address & MemoryRegion.OAM_MASK)

    def read_8_palram(self, address: int) -> int:
        return self.palram[address & MemoryRegion.PALRAM_MASK]

    def read_8_vram(self, address: int) -> int:
        mask = self._get_vram_address_mask(address)
        return self.vram[address & mask]

    def read_8_oam(self, address: int) -> int:
        return self.oam[address & MemoryRegion.OAM_MASK]

    def write_32_palram(self, address: int, value: int):
        array_write_32(self.palram, address & MemoryRegion.PALRAM_MASK, value)

    def write_32_vram(self, address: int, value: int):
        mask = self._get_vram_address_mask(address)
        array_write_32(self.vram, address & mask, value)

    def write_32_oam(self, address: int, value: int):
        array_write_32(self.oam, address & MemoryRegion.OAM_MASK, value)

    def write_16_palram(self, address: int, value: int):
        array_write_16(self.palram, address & MemoryRegion.PALRAM_MASK, value)

    def write_16_vram(self, address: int, value: int):
        mask = self._get_vram_address_mask(address)
        array_write_16(self.vram, address & mask, value)

    def write_16_oam(self, address: int, value: int):
        array_write_16(self.oam, address & MemoryRegion.OAM_MASK, value)

    def write_8_palram(self, address: int, value: int):
        # Writing a byte to PALRAM writes the value to both the upper and lower 8-bits
        # of the addressed halfword
        address = address & ~0b1  # Align address to 2-byte boundary
        self.palram[address & MemoryRegion.PALRAM_MASK] = value
        self.palram[(address + 1) & MemoryRegion.PALRAM_MASK] = value

    def write_8_vram(self, address: int, value: int):
        # VRAM is split into BG and OBJ regions.
        # Size of the BG region changes depending on whether we are in bitmap mode
        video_mode = self.display_control.video_mode
        bitmapped = video_mode in [VideoMode.MODE_3, VideoMode.MODE_4, VideoMode.MODE_5]
        bg_region_end = 0x14000 if bitmapped else 0x10000

        # Ignore attempts to write a byte into OBJ, but allow writes into BG
        if address & 0x1FFFF < bg_region_end:
            # Writing a byte to BG writes the value to both the upper and lower 8-bits
            # of the addressed halfword
            address = address & ~0b1  # Align address to 2-byte boundary
            mask_1 = self._get_vram_address_mask(address)
            mask_2 = self._get_vram_address_mask(address + 1)
            self.vram[address & mask_1] = value
            self.vram[(address + 1) & mask_2] = value

    def _get_vram_address_mask(self, address: int):
        """
        Get the appropriate mask for the given VRAM address.

        GBA VRAM is 96 KB (64K + 32K) but is mirrored in 128 KB steps
        (64K + 32K + 32K, the two 32K blocks being mirrors of each other).

        Hence, if the address is in the mirrored upper region (bit 16 set), we use
        VRAM_MASK_1 to wrap it back into valid VRAM space. Otherwise, we use VRAM_MASK_2.
        """
        return MemoryRegion.VRAM_MASK_1 if get_bit(address, 16) else MemoryRegion.VRAM_MASK_2



cpdef enum ThumbALUOpcode:
    ThumbALUOpcode_AND = 0
    ThumbALUOpcode_EOR = 1
    ThumbALUOpcode_LSL = 2
    ThumbALUOpcode_LSR = 3
    ThumbALUOpcode_ASR = 4
    ThumbALUOpcode_ADC = 5
    ThumbALUOpcode_SBC = 6
    ThumbALUOpcode_ROR = 7
    ThumbALUOpcode_TST = 8
    NEG = 9
    ThumbALUOpcode_CMP = 10
    ThumbALUOpcode_CMN = 11
    ThumbALUOpcode_ORR = 12
    MUL = 13
    ThumbALUOpcode_BIC = 14
    ThumbALUOpcode_MVN = 15




def arm_mrs(cpu: CPU, instr: int):
    """
    Execute an MRS instruction
    (transfer PSR contents to general-purpose register)
    """
    rd = get_bits(instr, 12, 15)
    psr = get_bit(instr, 22)
    cpu.regs.set(rd, cpu.regs.spsr.reg if psr else cpu.regs.cpsr.reg)
    cpu.advance_pc_arm()
    cpu.next_fetch_access = MemoryAccess.SEQUENTIAL

def arm_msr(cpu: CPU, instr: int):
    """
    Execute an MSR instruction
    (transfer general-purpose register contents or immediate value to PSR)
    """
    immediate = get_bit(instr, 25)
    if immediate:
        value = get_bits(instr, 0, 7)
        ror_amount = get_bits(instr, 8, 11) * 2
        value = ror_32(value, ror_amount)
    else:
        value = cpu.regs.get(get_bits(instr, 0, 3))
    write_to_flags_field = get_bit(instr, 19)
    write_to_status_field = get_bit(instr, 18)
    write_to_extension_field = get_bit(instr, 17)
    write_to_control_field = get_bit(instr, 16)
    flags_mask = 4278190080
    status_mask = 16711680
    extension_mask = 65280
    control_mask = 255
    mask = 0
    mask |= flags_mask if write_to_flags_field else 0
    mask |= status_mask if write_to_status_field else 0
    mask |= extension_mask if write_to_extension_field else 0
    mask |= control_mask if write_to_control_field else 0
    psr = get_bit(instr, 22)
    if psr:
        cpu.regs.spsr.reg = cpu.regs.spsr.reg & ~mask | value & mask
    else:
        if cpu.regs.cpsr.get_mode() == CPUMode.USER:
            mask &= flags_mask
        new_cpsr = cpu.regs.cpsr.reg & ~mask | value & mask
        new_cpsr_mode = new_cpsr & 31
        cpu.switch_mode(new_cpsr_mode)
        cpu.regs.cpsr.reg = new_cpsr
    cpu.advance_pc_arm()
    cpu.next_fetch_access = MemoryAccess.SEQUENTIAL


# ifndef CYTHON
from typing import Callable

from pyboy_advance.constants import EventTrigger
# endif

import heapq


class Event:
    def __init__(
        self,
        callback: Callable[[], None],
        delay: int,
        time: int,
        cancelled: bool,
    ):
        self.callback = callback
        self.delay = delay
        self.time = time
        self.cancelled = cancelled

    def __lt__(self, other):
        return self.time < other.time


class Scheduler:
    def __init__(self):
        self.cycles = 0
        self.events = []
        self.inactive_events = {}

    def schedule(
        self,
        callback: Callable[[], None],
        delay: int,
        trigger: EventTrigger,
    ) -> Event:
        event = Event(callback, delay, 0, False)
        if trigger == EventTrigger.IMMEDIATELY:
            event.time = self.cycles + delay
            heapq.heappush(self.events, event)
        else:
            if trigger not in self.inactive_events:
                self.inactive_events[trigger] = []
            self.inactive_events[trigger].append(event)
        return event

    def trigger(self, trigger: EventTrigger):
        if trigger not in self.inactive_events:
            self.inactive_events[trigger] = []

        while self.inactive_events[trigger]:
            event = self.inactive_events[trigger].pop()
            event.time = self.cycles + event.delay
            heapq.heappush(self.events, event)

    def idle(self, cycles: int):
        self.cycles += cycles
        self.process_events()

    def idle_until_next_event(self):
        if self.events:
            event = self.events[0]
            if event.time > self.cycles:
                self.idle(event.time - self.cycles)
        else:
            self.idle(1)

    def process_events(self):
        while self.events:
            event = self.events[0]
            if self.cycles >= event.time:
                heapq.heappop(self.events)
                if not event.cancelled:
                    # Rollback the cycle counter for the duration of the callback
                    delay = self.cycles - event.time
                    self.cycles -= delay

                    event.callback()

                    self.cycles += delay
            else:
                break



cimport cython

from libc.stdint cimport uint32_t, uint64_t

from pyboy_advance.cpu.cpu cimport CPU
from pyboy_advance.cpu.arm.constants cimport MultiplyLongOpcode
from pyboy_advance.memory.constants cimport MemoryAccess
from pyboy_advance.utils cimport get_bit, get_bits, sign_32, extend_sign_32

@cython.locals(mask=uint32_t)
cdef void arm_multiply(CPU, uint32_t) noexcept

@cython.locals(mask=uint32_t, rd_hi_val=uint64_t, rd_lo_val=uint64_t, rs_val=uint64_t, rm_val=uint64_t)
cdef void arm_multiply_long(CPU, uint32_t) noexcept

@cython.locals(mask=uint32_t)
cdef void arm_multiply_idle(CPU, uint32_t, bint) noexcept



# ifndef CYTHON
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pyboy_advance.cpu.cpu import CPU

from pyboy_advance.cpu.constants import ExceptionVector
# endif


def thumb_software_interrupt(cpu: CPU, instr: int):
    """Execute a THUMB.17 instruction (software interrupt)"""

    cpu.interrupt(ExceptionVector.SWI)



def thumb_move_shifted_register(cpu: CPU, instr: int):
    """Execute a THUMB.1 instruction (move shifted register)"""
    rs = get_bits(instr, 3, 5)
    rd = get_bits(instr, 0, 2)
    opcode = get_bits(instr, 11, 12)
    offset = get_bits(instr, 6, 10)
    result, carry = cpu.compute_shift(cpu.regs.get(rs), opcode, offset, True)
    cpu.regs.set(rd, result)
    cpu.regs.cpsr.set_sign_flag(sign_32(result))
    cpu.regs.cpsr.set_zero_flag(result == 0)
    cpu.regs.cpsr.set_carry_flag(carry)
    cpu.advance_pc_thumb()
    cpu.next_fetch_access = MemoryAccess.SEQUENTIAL

def thumb_add_subtract(cpu: CPU, instr: int):
    """Execute a THUMB.2 instruction (add/subtract)"""
    rs = get_bits(instr, 3, 5)
    rd = get_bits(instr, 0, 2)
    opcode = get_bits(instr, 9, 10)
    immediate_bit = get_bit(instr, 10)
    op1 = cpu.regs.get(rs)
    op2 = get_bits(instr, 6, 8) if immediate_bit else cpu.regs.get(get_bits(instr, 6, 8))
    if opcode == 0:
        arm_alu_add(cpu, op1, op2, rd, True)
    elif opcode == 1:
        arm_alu_sub(cpu, op1, op2, rd, True)
    elif opcode == 2:
        arm_alu_add(cpu, op1, op2, rd, True)
    elif opcode == 3:
        arm_alu_sub(cpu, op1, op2, rd, True)
    cpu.advance_pc_thumb()
    cpu.next_fetch_access = MemoryAccess.SEQUENTIAL

def thumb_move_compare_add_subtract(cpu: CPU, instr: int):
    """Execute a THUMB.3 instruction (move/compare/add/subtract immediate)"""
    rd = get_bits(instr, 8, 10)
    opcode = get_bits(instr, 11, 12)
    value = get_bits(instr, 0, 7)
    if opcode == 0:
        arm_alu_mov(cpu, value, rd, True, cpu.regs.cpsr.get_carry_flag())
    elif opcode == 1:
        arm_alu_cmp(cpu, cpu.regs.get(rd), value)
    elif opcode == 2:
        arm_alu_add(cpu, cpu.regs.get(rd), value, rd, True)
    elif opcode == 3:
        arm_alu_sub(cpu, cpu.regs.get(rd), value, rd, True)
    cpu.advance_pc_thumb()
    cpu.next_fetch_access = MemoryAccess.SEQUENTIAL

def thumb_alu(cpu: CPU, instr: int):
    """Execute a THUMB.4 instruction (miscellaneous ALU operations)"""
    rs = get_bits(instr, 3, 5)
    rd = get_bits(instr, 0, 2)
    op1 = cpu.regs.get(rd)
    op2 = cpu.regs.get(rs)
    opcode = get_bits(instr, 6, 9)
    if opcode == ThumbALUOpcode.ThumbALUOpcode_AND:
        arm_alu_and(cpu, op1, op2, rd, True, cpu.regs.cpsr.get_carry_flag())
    elif opcode == ThumbALUOpcode.ThumbALUOpcode_EOR:
        arm_alu_eor(cpu, op1, op2, rd, True, cpu.regs.cpsr.get_carry_flag())
    elif opcode == ThumbALUOpcode.ThumbALUOpcode_LSL:
        thumb_alu_shift(cpu, op1, op2, rd, ShiftType.LSL)
    elif opcode == ThumbALUOpcode.ThumbALUOpcode_LSR:
        thumb_alu_shift(cpu, op1, op2, rd, ShiftType.LSR)
    elif opcode == ThumbALUOpcode.ThumbALUOpcode_ASR:
        thumb_alu_shift(cpu, op1, op2, rd, ShiftType.ASR)
    elif opcode == ThumbALUOpcode.ThumbALUOpcode_ADC:
        arm_alu_adc(cpu, op1, op2, rd, True)
    elif opcode == ThumbALUOpcode.ThumbALUOpcode_SBC:
        arm_alu_sbc(cpu, op1, op2, rd, True)
    elif opcode == ThumbALUOpcode.ThumbALUOpcode_ROR:
        thumb_alu_shift(cpu, op1, op2, rd, ShiftType.ROR)
    elif opcode == ThumbALUOpcode.ThumbALUOpcode_TST:
        arm_alu_tst(cpu, op1, op2, cpu.regs.cpsr.get_carry_flag())
    elif opcode == ThumbALUOpcode.NEG:
        arm_alu_sub(cpu, 0, op2, rd, True)
    elif opcode == ThumbALUOpcode.ThumbALUOpcode_CMP:
        arm_alu_cmp(cpu, op1, op2)
    elif opcode == ThumbALUOpcode.ThumbALUOpcode_CMN:
        arm_alu_cmn(cpu, op1, op2)
    elif opcode == ThumbALUOpcode.ThumbALUOpcode_ORR:
        arm_alu_orr(cpu, op1, op2, rd, True, cpu.regs.cpsr.get_carry_flag())
    elif opcode == ThumbALUOpcode.MUL:
        thumb_alu_multiply(cpu, op1, op2, rd)
    elif opcode == ThumbALUOpcode.ThumbALUOpcode_BIC:
        arm_alu_bic(cpu, op1, op2, rd, True, cpu.regs.cpsr.get_carry_flag())
    elif opcode == ThumbALUOpcode.ThumbALUOpcode_MVN:
        arm_alu_mvn(cpu, op2, rd, True, cpu.regs.cpsr.get_carry_flag())
    cpu.advance_pc_thumb()
    cpu.next_fetch_access = MemoryAccess.SEQUENTIAL

def thumb_alu_shift(cpu: CPU, value: int, shift: int, rd: int, shift_type: ShiftType):
    shift &= 255
    result, carry = cpu.compute_shift(value, shift_type, shift, False)
    cpu.regs.set(rd, result)
    cpu.regs.cpsr.set_sign_flag(sign_32(result))
    cpu.regs.cpsr.set_zero_flag(result == 0)
    cpu.regs.cpsr.set_carry_flag(carry)
    cpu.scheduler.idle(1)

def thumb_alu_multiply(cpu: CPU, op1: int, op2: int, rd: int):
    mask = 4294967295
    result = op1 * op2 & mask
    cpu.regs.set(rd, result)
    cpu.regs.cpsr.set_sign_flag(sign_32(result))
    cpu.regs.cpsr.set_zero_flag(result == 0)
    arm_multiply_idle(cpu, op1, True)

def thumb_high_reg_branch_exchange(cpu: CPU, instr: int):
    """Execute a THUMB.5 instruction (high register operations or branch exchange)"""
    rs = get_bit(instr, 6) << 3 | get_bits(instr, 3, 5)
    rd = get_bit(instr, 7) << 3 | get_bits(instr, 0, 2)
    opcode = get_bits(instr, 8, 9)
    if opcode == 0:
        cpu.regs.set(rd, add_32(cpu.regs.get(rd), cpu.regs.get(rs)))
        if rd == cpu.regs.PC:
            cpu.flush_pipeline()
            return
    elif opcode == 1:
        arm_alu_cmp(cpu, cpu.regs.get(rd), cpu.regs.get(rs))
    elif opcode == 2:
        cpu.regs.set(rd, cpu.regs.get(rs))
        if rd == cpu.regs.PC:
            cpu.flush_pipeline()
            return
    elif opcode == 3:
        address = cpu.regs.get(rs)
        cpu.regs.set_pc(address & ~1)
        cpu.regs.cpsr.set_state(get_bit(address, 0))
        cpu.flush_pipeline()
        return
    cpu.advance_pc_thumb()
    cpu.next_fetch_access = MemoryAccess.SEQUENTIAL

def thumb_get_address(cpu: CPU, instr: int):
    """Execute a THUMB.12 instruction (get address relative to SP/PC)"""
    opcode = get_bit(instr, 11)
    rd = get_bits(instr, 8, 10)
    offset = get_bits(instr, 0, 7) * 4
    if opcode:
        cpu.regs.set(rd, add_32(cpu.regs.get_sp(), offset))
    else:
        cpu.regs.set(rd, add_32(cpu.regs.get_pc() & ~2, offset))
    cpu.advance_pc_thumb()
    cpu.next_fetch_access = MemoryAccess.SEQUENTIAL

def thumb_add_offset_to_stack_pointer(cpu: CPU, instr: int):
    """Execute a THUMB.13 instruction (add offset to SP)"""
    sign = -1 if get_bit(instr, 7) else 1
    offset = sign * get_bits(instr, 0, 6) * 4
    cpu.regs.set_sp(add_32(cpu.regs.get_sp(), offset))
    cpu.advance_pc_thumb()
    cpu.next_fetch_access = MemoryAccess.SEQUENTIAL


# ifndef CYTHON
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pyboy_advance.cpu.cpu import CPU

from pyboy_advance.cpu.constants import CPUMode
from pyboy_advance.memory.constants import MemoryAccess
from pyboy_advance.utils import get_bits, get_bit, ror_32
# endif


def arm_mrs(cpu: CPU, instr: int):
    """
    Execute an MRS instruction
    (transfer PSR contents to general-purpose register)
    """

    # Destination reg
    rd = get_bits(instr, 12, 15)

    # Source PSR; 0 = CPSR, 1 = SPSR_<current_mode>
    psr = get_bit(instr, 22)
    cpu.regs.set(rd, cpu.regs.spsr.reg if psr else cpu.regs.cpsr.reg)

    cpu.advance_pc_arm()
    cpu.next_fetch_access = MemoryAccess.SEQUENTIAL


def arm_msr(cpu: CPU, instr: int):
    """
    Execute an MSR instruction
    (transfer general-purpose register contents or immediate value to PSR)
    """

    immediate = get_bit(instr, 25)
    if immediate:
        value = get_bits(instr, 0, 7)
        ror_amount = get_bits(instr, 8, 11) * 2
        value = ror_32(value, ror_amount)
    else:
        value = cpu.regs.get(get_bits(instr, 0, 3))

    write_to_flags_field = get_bit(instr, 19)
    write_to_status_field = get_bit(instr, 18)
    write_to_extension_field = get_bit(instr, 17)
    write_to_control_field = get_bit(instr, 16)

    flags_mask = 0xFF000000
    status_mask = 0x00FF0000
    extension_mask = 0x0000FF00
    control_mask = 0x000000FF

    mask = 0
    mask |= flags_mask if write_to_flags_field else 0
    mask |= status_mask if write_to_status_field else 0
    mask |= extension_mask if write_to_extension_field else 0
    mask |= control_mask if write_to_control_field else 0

    # Destination PSR; 0 = CPSR, 1 = SPSR_<current_mode>
    psr = get_bit(instr, 22)

    if psr:
        cpu.regs.spsr.reg = (cpu.regs.spsr.reg & ~mask) | (value & mask)
    else:
        # In user mode, only the condition flags can be changed
        if cpu.regs.cpsr.mode == CPUMode.USER:
            mask &= flags_mask

        new_cpsr = (cpu.regs.cpsr.reg & ~mask) | (value & mask)
        new_cpsr_mode = new_cpsr & 0b11111
        cpu.switch_mode(new_cpsr_mode)
        cpu.regs.cpsr.reg = new_cpsr

    cpu.advance_pc_arm()
    cpu.next_fetch_access = MemoryAccess.SEQUENTIAL



from libc.stdint cimport uint32_t

from pyboy_advance.utils cimport get_bits, get_bit, set_bit

cdef class DisplayControlRegister:
    cdef uint32_t reg

    cdef int get_video_mode(self) noexcept
    cdef bint get_page_select(self) noexcept
    cdef bint get_hblank_interval_free(self) noexcept
    cdef bint get_obj_vram_dimension(self) noexcept
    cdef bint get_force_blank(self) noexcept
    cdef bint display_bg(self, int) noexcept
    cdef bint get_display_obj(self) noexcept
    cdef bint display_window(self, int) noexcept

cdef class DisplayStatusRegister:
    cdef uint32_t reg

    cdef bint get_vblank_status(self) noexcept
    cdef void set_vblank_status(self, bint) noexcept
    cdef bint get_hblank_status(self) noexcept
    cdef void set_hblank_status(self, bint) noexcept
    cdef bint get_vcount_trigger_status(self) noexcept
    cdef void set_vcount_trigger_status(self, bint) noexcept
    cdef bint get_vblank_irq(self) noexcept
    cdef void set_vblank_irq(self, bint) noexcept
    cdef bint get_hblank_irq(self) noexcept
    cdef void set_hblank_irq(self, bint) noexcept
    cdef bint get_vcount_irq(self) noexcept
    cdef void set_vcount_irq(self, bint) noexcept
    cdef uint32_t get_vcount_trigger_value(self) noexcept
    cdef void set_vcount_trigger_value(self, uint32_t) noexcept

cdef class BackgroundControlRegister:
    cdef uint32_t reg

    cdef uint32_t get_priority(self) noexcept
    cdef uint32_t get_tile_set_block(self) noexcept
    cdef bint get_mosaic(self) noexcept
    cdef bint get_colour_256(self) noexcept
    cdef uint32_t get_tile_map_block(self) noexcept
    cdef bint get_wraparound(self) noexcept
    cdef uint32_t get_size(self) noexcept

cdef class WindowControlRegister:
    cdef uint32_t reg

    cdef bint display_layer(self, int) noexcept
    cdef bint get_enable_blending(self) noexcept

cdef class BlendControlRegister:
    cdef uint32_t reg

    cdef bint blend_source(self, int) noexcept
    cdef bint blend_target(self, int) noexcept
    cdef int get_blend_mode(self) noexcept

cdef class BlendAlphaRegister:
    cdef uint32_t reg

    cdef uint32_t get_coefficient_source(self) noexcept
    cdef uint32_t get_coefficient_target(self) noexcept

cdef class BlendBrightnessRegister:
    cdef uint32_t reg

    cdef uint32_t get_brightness(self) noexcept



import ast
import os
import re
from dataclasses import dataclass
from collections import Counter


@dataclass
class Constant:
    scope: str | None
    name: str
    replacement_name: str | None
    value: int


def preprocess_constants(file_paths: list[str | os.PathLike]) -> list[str | os.PathLike]:
    """
    Generate constants.pxd and constants.pyx files for constants.py files containing
    enums and global constants.

    IntEnum/IntFlag/Enum definitions are converted to named cpdef enum definitions.
    Global constants are placed in an anonymous cdef enum.

    Name mangling is applied to avoid name conflicts (in C, all enum members share
    the same namespace).
    """
    preprocessed = []
    constants: dict[tuple[str | None, str], Constant] = {}
    constant_names: set[str] = set()

    for file_path in file_paths:
        file_name = os.path.basename(file_path)
        base, ext = os.path.splitext(file_name)
        if "constants" in base:
            generated_pyx_file = generate_constants_pxd_pyx(
                file_path,
                constants,
                constant_names,
            )
            preprocessed.append(generated_pyx_file)
        else:
            preprocessed.append(file_path)

    insert_replacements(preprocessed, constants)

    return preprocessed


def generate_constants_pxd_pyx(
    py_file_path: str | os.PathLike,
    constants: dict[tuple[str | None, str], Constant],
    constant_names: set[str],
) -> str | os.PathLike:
    with open(py_file_path, "r", encoding="utf-8") as f:
        content = f.read()

    tree = ast.parse(content)

    pxd_lines = []
    global_constants = []

    for node in tree.body:
        if isinstance(node, ast.ClassDef):
            # Convert IntEnum/IntFlag/Enum classes to cpdef enum definitions
            bases = [base.id for base in node.bases if isinstance(base, ast.Name)]
            if not any(base in ("IntEnum", "IntFlag", "Enum") for base in bases):
                continue

            pxd_lines.append(f"cpdef enum {node.name}:")

            for item in node.body:
                if isinstance(item, ast.Assign):
                    if len(item.targets) == 1 and isinstance(item.targets[0], ast.Name):
                        member_name = item.targets[0].id
                        member_value = _eval_int_expr(item.value, constants, node.name)

                        if member_name in constant_names:
                            new_member_name = node.name + "_" + member_name
                            constants[(node.name, member_name)] = Constant(
                                node.name,
                                member_name,
                                new_member_name,
                                member_value,
                            )
                            constant_names.add(new_member_name)
                            pxd_lines.append(f"    {new_member_name} = {member_value}")
                        else:
                            constants[(node.name, member_name)] = Constant(
                                node.name,
                                member_name,
                                None,
                                member_value,
                            )
                            constant_names.add(member_name)
                            pxd_lines.append(f"    {member_name} = {member_value}")

            pxd_lines.append("")

        elif isinstance(node, ast.Assign):
            # Collect global constants
            if len(node.targets) == 1 and isinstance(node.targets[0], ast.Name):
                constant_name = node.targets[0].id
                constant_value = _eval_int_expr(node.value, constants, None)

                if constant_name in constant_names:
                    new_constant_name = "GLOB_" + constant_name
                    constants[(None, constant_name)] = Constant(
                        None,
                        constant_name,
                        new_constant_name,
                        constant_value,
                    )
                    constant_names.add(new_constant_name)
                else:
                    constants[(None, constant_name)] = Constant(
                        None,
                        constant_name,
                        None,
                        constant_value,
                    )
                    constant_names.add(constant_name)

                global_constants.append(constants[(None, constant_name)])

    if global_constants:
        pxd_lines.append("cdef enum:")
        for constant in global_constants:
            if constant.replacement_name:
                pxd_lines.append(f"    {constant.replacement_name} = {constant.value}")
            else:
                pxd_lines.append(f"    {constant.name} = {constant.value}")
        pxd_lines.append("")

    base = py_file_path.split(".", 1)[0]

    pxd_path = base + ".pxd"
    with open(pxd_path, "w", encoding="utf-8") as f:
        f.write("\n".join(pxd_lines) + "\n")

    pyx_path = base + ".pyx"
    with open(pyx_path, "w", encoding="utf-8") as f:
        f.write("")

    return pyx_path


def insert_replacements(
    file_paths: list[str | os.PathLike],
    constants: dict[tuple[str | None, str], Constant],
):
    replacements: dict[str, str] = {}
    enum_lens = Counter()

    for constant in constants.values():
        if constant.replacement_name:
            if constant.scope:
                pattern = rf"\b{constant.scope}\.{constant.name}\b"
                replacements[pattern] = f"{constant.scope}.{constant.replacement_name}"
            else:
                pattern = rf"[^.]{constant.name}\b"
                replacements[pattern] = constant.replacement_name

        if constant.scope:
            enum_lens[constant.scope] += 1

    for enum_name, enum_len in enum_lens.items():
        pattern = f"len\\({enum_name}\\)"
        replacements[pattern] = str(enum_len)

    for file_path in file_paths:
        with open(file_path, "r", encoding="utf-8") as f:
            content = f.read()

            for pattern, replacement in replacements.items():
                content = re.sub(pattern, replacement, content)

        with open(file_path, "w", encoding="utf-8") as f:
            f.write(content)


def _eval_int_expr(
    node: ast.AST, constants: dict[tuple[str | None, str], Constant], scope: str | None
) -> int:
    if isinstance(node, ast.Constant) and isinstance(node.value, int):
        return node.value
    if isinstance(node, ast.Name):
        if (scope, node.id) in constants:
            return constants[(scope, node.id)].value
        if (None, node.id) in constants:
            return constants[(None, node.id)].value
        raise ValueError(f"Unknown name: ({scope}, {node.id})")
    if isinstance(node, ast.BinOp):
        left = _eval_int_expr(node.left, constants, scope)
        right = _eval_int_expr(node.right, constants, scope)
        op = node.op
        if isinstance(op, ast.Add):
            return left + right
        if isinstance(op, ast.Sub):
            return left - right
        if isinstance(op, ast.Mult):
            return left * right
        if isinstance(op, ast.FloorDiv):
            return left // right
        if isinstance(op, ast.Mod):
            return left % right
        if isinstance(op, ast.LShift):
            return left << right
        if isinstance(op, ast.RShift):
            return left >> right
        if isinstance(op, ast.BitOr):
            return left | right
        if isinstance(op, ast.BitAnd):
            return left & right
        if isinstance(op, ast.BitXor):
            return left ^ right
        raise ValueError(f"Unsupported binary operator: {type(op).__name__}")
    if isinstance(node, ast.UnaryOp):
        operand = _eval_int_expr(node.operand, constants, scope)
        if isinstance(node.op, ast.USub):
            return -operand
        if isinstance(node.op, ast.UAdd):
            return +operand
        if isinstance(node.op, ast.Invert):
            return ~operand
        raise ValueError(f"Unsupported unary operator: {type(node.op).__name__}")
    raise ValueError(f"Unsupported expression node: {type(node).__name__}")



from libc.stdint cimport uint32_t

from pyboy_advance.cpu.cpu cimport CPU
from pyboy_advance.utils cimport extend_sign_24, get_bit, add_32

cdef void arm_branch(CPU, uint32_t) noexcept
cdef void arm_branch_exchange(CPU, uint32_t) noexcept



class Keypad:

    def __init__(self, interrupt_controller: InterruptController):
        self.interrupt_controller = interrupt_controller
        self.key_input = 1023
        self.key_control = KeypadInterruptControlRegister()

    def press_key(self, key: Key):
        self.key_input &= ~key
        if self.key_control.get_irq_enable() and self._evaluate_irq_condition():
            self.interrupt_controller.signal(Interrupt.KEYPAD)

    def release_key(self, key: Key):
        self.key_input |= key
        if self.key_control.get_irq_enable() and self._evaluate_irq_condition():
            self.interrupt_controller.signal(Interrupt.KEYPAD)

    def process_window_event(self, event: WindowEvent):
        if event == WindowEvent.PRESS_BUTTON_A:
            self.press_key(Key.BUTTON_A)
        elif event == WindowEvent.PRESS_BUTTON_B:
            self.press_key(Key.BUTTON_B)
        elif event == WindowEvent.PRESS_BUTTON_SELECT:
            self.press_key(Key.BUTTON_SELECT)
        elif event == WindowEvent.PRESS_BUTTON_START:
            self.press_key(Key.BUTTON_START)
        elif event == WindowEvent.PRESS_DPAD_RIGHT:
            self.press_key(Key.DPAD_RIGHT)
        elif event == WindowEvent.PRESS_DPAD_LEFT:
            self.press_key(Key.DPAD_LEFT)
        elif event == WindowEvent.PRESS_DPAD_UP:
            self.press_key(Key.DPAD_UP)
        elif event == WindowEvent.PRESS_DPAD_DOWN:
            self.press_key(Key.DPAD_DOWN)
        elif event == WindowEvent.PRESS_SHOULDER_RIGHT:
            self.press_key(Key.SHOULDER_RIGHT)
        elif event == WindowEvent.PRESS_SHOULDER_LEFT:
            self.press_key(Key.SHOULDER_LEFT)
        elif event == WindowEvent.RELEASE_BUTTON_A:
            self.release_key(Key.BUTTON_A)
        elif event == WindowEvent.RELEASE_BUTTON_B:
            self.release_key(Key.BUTTON_B)
        elif event == WindowEvent.RELEASE_BUTTON_SELECT:
            self.release_key(Key.BUTTON_SELECT)
        elif event == WindowEvent.RELEASE_BUTTON_START:
            self.release_key(Key.BUTTON_START)
        elif event == WindowEvent.RELEASE_DPAD_RIGHT:
            self.release_key(Key.DPAD_RIGHT)
        elif event == WindowEvent.RELEASE_DPAD_LEFT:
            self.release_key(Key.DPAD_LEFT)
        elif event == WindowEvent.RELEASE_DPAD_UP:
            self.release_key(Key.DPAD_UP)
        elif event == WindowEvent.RELEASE_DPAD_DOWN:
            self.release_key(Key.DPAD_DOWN)
        elif event == WindowEvent.RELEASE_SHOULDER_RIGHT:
            self.release_key(Key.SHOULDER_RIGHT)
        elif event == WindowEvent.RELEASE_SHOULDER_LEFT:
            self.release_key(Key.SHOULDER_LEFT)

    def _evaluate_irq_condition(self) -> bool:
        if self.key_control.get_irq_if_all():
            return ~self.key_input & self.key_control.get_key_select() == self.key_control.get_key_select()
        else:
            return ~self.key_input & self.key_control.get_key_select() != 0

class KeypadInterruptControlRegister:

    def __init__(self):
        self.reg = 0

    def get_key_select(self) -> int:
        return get_bits(self.reg, 0, 9)

    def get_irq_enable(self) -> bint:
        return get_bit(self.reg, 14)

    def get_irq_if_all(self) -> bint:
        return get_bit(self.reg, 15)


import argparse
import sys

from pyboy_advance import PyBoyAdvance


def main():
    parser = argparse.ArgumentParser("pyboy_advance")
    parser.add_argument("rom", type=str)
    parser.add_argument("--bios", type=str)
    parser.add_argument("--skip-bios", action="store_true")
    parser.add_argument("--emulation-speed", type=float, default=1)
    args = parser.parse_args()

    pyboy_advance = PyBoyAdvance(
        args.rom,
        args.bios,
        skip_bios=args.skip_bios,
        emulation_speed=args.emulation_speed,
    )

    pyboy_advance.run()


if __name__ == "__main__":
    sys.exit(main())



# ifndef CYTHON
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pyboy_advance.cpu.cpu import CPU

from pyboy_advance.memory.constants import MemoryAccess
from pyboy_advance.utils import (
    get_bit,
    get_bits,
    add_32,
)
# endif


def thumb_multiple_load_store(cpu: CPU, instr: int):
    """Execute a THUMB.15 instruction (multiple load/store)"""

    reg_list = get_bits(instr, 0, 7)
    base_reg = get_bits(instr, 8, 10)
    address = cpu.regs.get(base_reg)

    cpu.advance_pc_thumb()
    # cpu.prefetch_access_type = MemoryAccess.NON_SEQUENTIAL

    load = get_bit(instr, 11)  # 0 = STMIA (Store), 1 = LDMIA (Load)

    if reg_list == 0:
        if load:
            cpu.regs.pc = cpu.memory.read_32(cpu.regs.get(base_reg), MemoryAccess.NON_SEQUENTIAL)
            cpu.flush_pipeline()
        else:
            cpu.memory.write_32(cpu.regs.get(base_reg), cpu.regs.pc, MemoryAccess.NON_SEQUENTIAL)
        cpu.regs.set(base_reg, add_32(cpu.regs.get(base_reg), 0x40))
        return

    # Avoiding usage of sum() here to optimize Cython output
    count = 0
    for reg in range(8):
        count += get_bit(instr, reg)
    count *= 4

    access_type = MemoryAccess.NON_SEQUENTIAL

    if load:
        cpu.regs.set(base_reg, add_32(cpu.regs.get(base_reg), count))
        cpu.scheduler.idle(1)

        for reg in range(8):
            if get_bit(reg_list, reg):
                cpu.regs.set(reg, cpu.memory.read_32(address, access_type))
                address = add_32(address, 4)
                access_type = MemoryAccess.SEQUENTIAL
    else:
        first = True
        for reg in range(8):
            if get_bit(reg_list, reg):
                cpu.memory.write_32(address, cpu.regs.get(reg), access_type)
                address = add_32(address, 4)
                access_type = MemoryAccess.SEQUENTIAL
                if first:
                    cpu.regs.set(base_reg, add_32(cpu.regs.get(base_reg), count))
                    first = False


def thumb_push_pop_registers(cpu: CPU, instr: int):
    """Execute a THUMB.14 instruction (push/pop registers to the stack)"""

    reg_list = get_bits(instr, 0, 8)
    opcode = get_bit(instr, 11)

    cpu.advance_pc_thumb()
    # cpu.prefetch_access_type = MemoryAccess.NON_SEQUENTIAL

    if opcode:
        thumb_pop_registers(cpu, reg_list)
    else:
        thumb_push_registers(cpu, reg_list)


def thumb_push_registers(cpu: CPU, reg_list: int):
    """
    Push the specified registers to the stack.

    :param cpu:      The CPU executing this instruction.
    :param reg_list: A 9-bit wide bitfield specifying which registers to push.
                     Bits 0-7 correspond to registers R0-R7 and bit 8 corresponds
                     to LR (R14).
    """

    if reg_list == 0:
        cpu.regs.sp = add_32(cpu.regs.sp, -0x40)
        cpu.memory.write_32(cpu.regs.sp, cpu.regs.pc, MemoryAccess.NON_SEQUENTIAL)
        return

    push_lr = get_bit(reg_list, 8)
    if push_lr:
        cpu.regs.sp = add_32(cpu.regs.sp, -4)
        cpu.memory.write_32(cpu.regs.sp, cpu.regs.lr, MemoryAccess.NON_SEQUENTIAL)

    for reg in range(7, -1, -1):
        if get_bit(reg_list, reg):
            cpu.regs.sp = add_32(cpu.regs.sp, -4)
            cpu.memory.write_32(cpu.regs.sp, cpu.regs.get(reg), MemoryAccess.SEQUENTIAL)


def thumb_pop_registers(cpu: CPU, reg_list: int):
    """
    Pop the specified registers off the stack.

    :param cpu:      The CPU executing this instruction.
    :param reg_list: A 9-bit wide bitfield specifying which registers to pop.
                     Bits 0-7 correspond to registers R0-R7 and bit 8 corresponds
                     to PC (R15).
    """

    if reg_list == 0:
        cpu.regs.pc = cpu.memory.read_32(cpu.regs.sp, MemoryAccess.NON_SEQUENTIAL)
        cpu.regs.sp = add_32(cpu.regs.sp, 0x40)
        cpu.flush_pipeline()
        return

    access_type = MemoryAccess.NON_SEQUENTIAL
    for reg in range(8):
        if get_bit(reg_list, reg):
            cpu.regs.set(reg, cpu.memory.read_32(cpu.regs.sp, access_type))
            cpu.regs.sp = add_32(cpu.regs.sp, 4)
            access_type = MemoryAccess.SEQUENTIAL

    pop_pc = get_bit(reg_list, 8)
    if pop_pc:
        cpu.regs.pc = cpu.memory.read_32(cpu.regs.sp, access_type)
        cpu.regs.sp = add_32(cpu.regs.sp, 4)
        cpu.flush_pipeline()

    cpu.scheduler.idle(1)



# ifndef CYTHON
from array import array
from typing import TypeAlias

bint: TypeAlias = int
"""
This type alias is used to indicate that the value is intended to behave
like a boolean (0 for False, 1 for True) while still being an integer.
Casting to bool has a measurable performance hit, so it is better to avoid
casts and use bint where possible.
"""
# endif


def get_bit(num: int, i: int) -> bint:
    return (num >> i) & 1


def get_bits(num: int, start: int, end: int) -> int:
    num_bits = end - start + 1
    mask = (1 << num_bits) - 1
    return (num >> start) & mask


def set_bit(num: int, i: int, bit: bint) -> int:
    return (num & ~(1 << i)) | (bit << i)


def sign_32(num: int) -> bint:
    return get_bit(num, 31)


def sign_24(num: int) -> bint:
    return get_bit(num, 23)


def sign_23(num: int) -> bint:
    return get_bit(num, 22)


def sign_16(num: int) -> bint:
    return get_bit(num, 15)


def sign_12(num: int) -> bint:
    return get_bit(num, 11)


def sign_9(num: int) -> bint:
    return get_bit(num, 8)


def sign_8(num: int) -> bint:
    return get_bit(num, 7)


def extend_sign_32(num: int) -> int:
    extension = 0xFFFFFFFF00000000
    return num | extension if sign_32(num) else num


def extend_sign_24(num: int) -> int:
    extension = 0xFF000000
    return num | extension if sign_24(num) else num


def extend_sign_23(num: int) -> int:
    extension = 0xFF800000
    return num | extension if sign_23(num) else num


def extend_sign_16(num: int) -> int:
    extension = 0xFFFF0000
    return num | extension if sign_16(num) else num


def extend_sign_12(num: int) -> int:
    extension = 0xFFFFF000
    return num | extension if sign_12(num) else num


def extend_sign_9(num: int) -> int:
    extension = 0xFFFFFE00
    return num | extension if sign_9(num) else num


def extend_sign_8(num: int) -> int:
    extension = 0xFFFFFF00
    return num | extension if sign_8(num) else num


def add_32(op1: int, op2: int) -> int:
    mask = 0xFFFFFFFF
    return (op1 + op2) & mask


def ror_32(num: int, amount: int) -> int:
    """
    Rotate right (ROR)

    :param num: 32-bit unsigned integer to rotate
    :param amount: number of bits to rotate to the right (must be <= 32)
    :return: num rotated right by amount bits
    """
    mask = 0xFFFFFFFF
    return ((num >> amount) | (num << (32 - amount))) & mask


def array_read_32(arr: array, address: int) -> int:
    """Read a little-endian 32-bit value from an array of bytes"""
    b0 = arr[address]
    b1 = arr[address + 1]
    b2 = arr[address + 2]
    b3 = arr[address + 3]
    return b0 | (b1 << 8) | (b2 << 16) | (b3 << 24)


def array_read_16(arr: array, address: int) -> int:
    """Read a little-endian 16-bit value from an array of bytes"""
    b0 = arr[address]
    b1 = arr[address + 1]
    return b0 | (b1 << 8)


def array_write_32(arr: array, address: int, value: int):
    """Write a 32-bit value to the given array of bytes in little-endian format"""
    arr[address] = value & 0xFF
    arr[address + 1] = (value >> 8) & 0xFF
    arr[address + 2] = (value >> 16) & 0xFF
    arr[address + 3] = (value >> 24) & 0xFF


def array_write_16(arr: array, address: int, value: int):
    """Write a 16-bit value to the given array of bytes in little-endian format"""
    arr[address] = value & 0xFF
    arr[address + 1] = (value >> 8) & 0xFF



from enum import IntEnum


class CPUState(IntEnum):
    ARM = 0
    THUMB = 1


class CPUMode(IntEnum):
    USER = 0b10000
    FIQ = 0b10001
    IRQ = 0b10010
    SWI = 0b10011
    ABORT = 0b10111
    UNDEFINED = 0b11011
    SYSTEM = 0b11111


class Condition(IntEnum):
    EQ = 0x0
    NE = 0x1
    HS = 0x2
    LO = 0x3
    MI = 0x4
    PL = 0x5
    VS = 0x6
    VC = 0x7
    HI = 0x8
    LS = 0x9
    GE = 0xA
    LT = 0xB
    GT = 0xC
    LE = 0xD
    AL = 0xE
    NV = 0xF


class ShiftType(IntEnum):
    LSL = 0
    LSR = 1
    ASR = 2
    ROR = 3


class ExceptionVector(IntEnum):
    RESET = 0x00
    UNDEFINED_INSTRUCTION = 0x04
    SWI = 0x08
    PREFETCH_ABORT = 0x0C
    DATA_ABORT = 0x10
    ADDRESS_EXCEEDS_26_BITS = 0x14
    IRQ = 0x18
    FIQ = 0x1C


class BankIndex(IntEnum):
    BANK_SYSTEM_USER = 0
    BANK_FIQ = 1
    BANK_IRQ = 2
    BANK_SWI = 3
    BANK_ABORT = 4
    BANK_UNDEFINED = 5


ARM_PC_INCREMENT = 4
THUMB_PC_INCREMENT = 2



from enum import IntEnum


class ALUOpcode(IntEnum):
    AND = 0x0
    EOR = 0x1
    SUB = 0x2
    RSB = 0x3
    ADD = 0x4
    ADC = 0x5
    SBC = 0x6
    RSC = 0x7
    TST = 0x8
    TEQ = 0x9
    CMP = 0xA
    CMN = 0xB
    ORR = 0xC
    MOV = 0xD
    BIC = 0xE
    MVN = 0xF


class DataTransferOpcode(IntEnum):
    STRH = 1  # Store halfword
    LDRD = 2  # Load doubleword
    STRD = 3  # Store doubleword

    LDRH = 1  # Load unsigned halfword
    LDRSB = 2  # Load signed byte
    LDRSH = 3  # Load signed halfword


class MultiplyLongOpcode(IntEnum):
    UMULL = 0b00
    UMLAL = 0b01
    SMULL = 0b10
    SMLAL = 0b11



cimport cython

from pyboy_advance.constants cimport EventTrigger

cdef class Event:
    cdef public object callback
    cdef public int delay
    cdef public int time
    cdef public bint cancelled

cdef class Scheduler:
    cdef int cycles
    cdef list events
    cdef dict inactive_events

    @cython.locals(event=Event)
    cdef Event schedule(self, object, int, int) noexcept

    @cython.locals(event=Event)
    cdef void trigger(self, int) noexcept

    cdef void idle(self, int) noexcept

    @cython.locals(event=Event)
    cdef void idle_until_next_event(self) noexcept

    @cython.locals(event=Event)
    cdef void process_events(self) noexcept



from libc.stdint cimport uint32_t

from pyboy_advance.cpu.cpu cimport CPU, InstrHandler, InstrPattern
from pyboy_advance.cpu.thumb.alu cimport (
    thumb_move_shifted_register,
    thumb_add_subtract,
    thumb_move_compare_add_subtract,
    thumb_alu,
    thumb_high_reg_branch_exchange,
    thumb_add_offset_to_stack_pointer,
    thumb_get_address,
)
from pyboy_advance.cpu.thumb.bdt cimport (
    thumb_multiple_load_store,
    thumb_push_pop_registers,
)
from pyboy_advance.cpu.thumb.branch cimport (
    thumb_unconditional_branch,
    thumb_conditional_branch,
    thumb_long_branch_with_link,
)
from pyboy_advance.cpu.thumb.sdt cimport (
    thumb_pc_relative_load,
    thumb_load_store_sign_extended,
    thumb_load_store_register_offset,
    thumb_load_store_immediate_offset,
    thumb_load_store_halfword,
    thumb_sp_relative_load_store,
)
from pyboy_advance.cpu.thumb.swi cimport thumb_software_interrupt

cdef InstrPattern[19] THUMB_PATTERNS

cdef InstrHandler[4096] THUMB_LUT

cdef void fill_thumb_lut() noexcept

cdef InstrHandler thumb_decoder(uint32_t) noexcept



class InterruptController:

    def __init__(self, scheduler: Scheduler):
        self.WRITE_INTERRUPT_REGISTERS_DELAY = 1
        self.UPDATE_IRQ_LINE_DELAY = 2
        self.scheduler = scheduler
        self._interrupt_enable = 0
        self._interrupt_flags = 0
        self._interrupt_master_enable = 0
        self._pending_interrupt_enable = 0
        self._pending_interrupt_flags = 0
        self._pending_interrupt_master_enable = 0
        self.irq_line = False
        self.power_down_mode = PowerDownMode.NONE

    def get_interrupt_enable(self) -> int:
        return self._interrupt_enable

    def set_interrupt_enable(self, value: int):
        self._pending_interrupt_enable = value & Interrupt.ALL
        self._schedule_write_interrupt_registers()

    def get_interrupt_flags(self) -> int:
        return self._interrupt_flags

    def set_interrupt_flags(self, value: int):
        self._pending_interrupt_flags &= ~value
        self._schedule_write_interrupt_registers()

    def get_interrupt_master_enable(self) -> bint:
        return self._interrupt_master_enable

    def set_interrupt_master_enable(self, value: bint):
        self._pending_interrupt_master_enable = value & 1
        self._schedule_write_interrupt_registers()

    def signal(self, interrupt: Interrupt):
        self._pending_interrupt_flags |= interrupt
        self._schedule_write_interrupt_registers()

    def _schedule_write_interrupt_registers(self):
        self.scheduler.schedule(self._write_interrupt_registers, self.WRITE_INTERRUPT_REGISTERS_DELAY, EventTrigger.IMMEDIATELY)

    def _write_interrupt_registers(self):
        self._interrupt_enable = self._pending_interrupt_enable
        self._interrupt_flags = self._pending_interrupt_flags
        self._interrupt_master_enable = self._pending_interrupt_master_enable
        interrupt_available = self._interrupt_enable & self._interrupt_flags
        if interrupt_available and self.power_down_mode == PowerDownMode.HALT:
            self.power_down_mode = PowerDownMode.NONE
        new_irq_line = interrupt_available and self._interrupt_master_enable
        if new_irq_line != self.irq_line:
            if new_irq_line:
                self.scheduler.schedule(self._set_irq_line, self.UPDATE_IRQ_LINE_DELAY, EventTrigger.IMMEDIATELY)
            else:
                self.scheduler.schedule(self._reset_irq_line, self.UPDATE_IRQ_LINE_DELAY, EventTrigger.IMMEDIATELY)

    def _set_irq_line(self):
        self.irq_line = True

    def _reset_irq_line(self):
        self.irq_line = False


# ifndef CYTHON
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pyboy_advance.cpu.cpu import CPU

from pyboy_advance.memory.constants import MemoryAccess
from pyboy_advance.utils import get_bit, get_bits, add_32
# endif


def arm_single_data_transfer(cpu: CPU, instr: int):
    """Execute a Single Data Transfer instruction (LDR, STR)"""

    rn = get_bits(instr, 16, 19)  # Base reg
    rd = get_bits(instr, 12, 15)  # Source/dest reg

    immediate_bit = get_bit(instr, 25)
    pre_post_bit = get_bit(instr, 24)
    up_down_bit = get_bit(instr, 23)
    byte_word_bit = get_bit(instr, 22)
    write_back_bit = get_bit(instr, 21)
    load_store_bit = get_bit(instr, 20)

    if immediate_bit:
        rm = get_bits(instr, 0, 3)
        shift = get_bits(instr, 4, 11)
        offset, _ = cpu.decode_and_compute_shift(cpu.regs.get(rm), shift)
    else:
        offset = get_bits(instr, 0, 11)

    base = cpu.regs.get(rn)
    address = add_32(base, offset if up_down_bit else -offset)
    effective_address = address if pre_post_bit else base

    cpu.advance_pc_arm()
    cpu.next_fetch_access = MemoryAccess.NON_SEQUENTIAL

    if load_store_bit:  # Load
        if byte_word_bit:
            # Load byte
            value = cpu.memory.read_8(effective_address, MemoryAccess.NON_SEQUENTIAL)
        else:
            # Load word
            value = cpu.memory.read_32_ror(effective_address, MemoryAccess.NON_SEQUENTIAL)

        # When post-indexing, write back is always enabled
        # When pre-indexing, write back is optional (controlled by W bit)
        if not pre_post_bit or write_back_bit:
            cpu.regs.set(rn, address)

        cpu.regs.set(rd, value)

        cpu.scheduler.idle(1)

        if rd == cpu.regs.PC:
            cpu.flush_pipeline()

    else:  # Store
        if byte_word_bit:
            # Store byte
            cpu.memory.write_8(effective_address, cpu.regs.get(rd), MemoryAccess.NON_SEQUENTIAL)
        else:
            # Store word
            cpu.memory.write_32(effective_address, cpu.regs.get(rd), MemoryAccess.NON_SEQUENTIAL)

        # When post-indexing, write back is always enabled
        # When pre-indexing, write back is optional (controlled by W bit)
        if not pre_post_bit or write_back_bit:
            cpu.regs.set(rn, address)



def arm_halfword_data_transfer(cpu: CPU, instr: int):
    """Execute a Halfword Data Transfer instruction (LDRH, STRH, etc.)"""
    rn = get_bits(instr, 16, 19)
    rd = get_bits(instr, 12, 15)
    pre_post_bit = get_bit(instr, 24)
    up_down_bit = get_bit(instr, 23)
    immediate_bit = get_bit(instr, 22)
    write_back_bit = get_bit(instr, 21)
    load_store_bit = get_bit(instr, 20)
    offset = get_bits(instr, 8, 11) << 4 | get_bits(instr, 0, 3) if immediate_bit else cpu.regs.get(get_bits(instr, 0, 3))
    base = cpu.regs.get(rn)
    address = add_32(base, offset if up_down_bit else -offset)
    effective_address = address if pre_post_bit else base
    cpu.advance_pc_arm()
    cpu.next_fetch_access = MemoryAccess.NON_SEQUENTIAL
    opcode = get_bits(instr, 5, 6)
    if load_store_bit:
        if not pre_post_bit or write_back_bit:
            cpu.regs.set(rn, address)
        if opcode == DataTransferOpcode.LDRH:
            cpu.regs.set(rd, cpu.memory.read_16_ror(effective_address, MemoryAccess.NON_SEQUENTIAL))
        elif opcode == DataTransferOpcode.LDRSB:
            cpu.regs.set(rd, cpu.memory.read_8_signed(effective_address, MemoryAccess.NON_SEQUENTIAL))
        elif opcode == DataTransferOpcode.LDRSH:
            cpu.regs.set(rd, cpu.memory.read_16_signed(effective_address, MemoryAccess.NON_SEQUENTIAL))
        cpu.scheduler.idle(1)
        if rd == cpu.regs.PC:
            cpu.flush_pipeline()
    else:
        if opcode == DataTransferOpcode.STRH:
            cpu.memory.write_16(effective_address, cpu.regs.get(rd), MemoryAccess.NON_SEQUENTIAL)
        elif opcode == DataTransferOpcode.LDRD:
            pass
        elif opcode == DataTransferOpcode.STRD:
            pass
        if not pre_post_bit or write_back_bit:
            cpu.regs.set(rn, address)


from libc.stdint cimport uint8_t, uint32_t
from cpython.array cimport array

from pyboy_advance.memory.constants cimport MemoryRegion
from pyboy_advance.utils cimport array_read_32, array_read_16

cdef class GamePak:
    cdef uint8_t[:] rom

    cdef uint32_t read_32(self, uint32_t) noexcept
    cdef uint32_t read_16(self, uint32_t) noexcept
    cdef uint32_t read_8(self, uint32_t) noexcept
